#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from collections.abc import Callable, Iterable
from copy import deepcopy
from datetime import datetime
from functools import wraps
from logging import Logger
from typing import Any, NamedTuple

import numpy as np
import numpy.typing as npt
from ax.core.arm import Arm
from ax.core.base_trial import BaseTrial, TrialStatus
from ax.core.batch_trial import BatchTrial
from ax.core.data import Data
from ax.core.experiment import Experiment
from ax.core.generator_run import GeneratorRun
from ax.core.map_metric import MapMetric
from ax.core.objective import MultiObjective
from ax.core.observation import ObservationFeatures
from ax.core.optimization_config import OptimizationConfig
from ax.core.trial import Trial
from ax.core.types import ComparisonOp
from ax.exceptions.core import AxError, UserInputError
from ax.utils.common.constants import Keys
from ax.utils.common.logger import get_logger
from pyre_extensions import none_throws

logger: Logger = get_logger(__name__)
TArmTrial = tuple[str, int]

# Threshold for switching to pending points extraction based on trial status.
MANY_TRIALS_IN_EXPERIMENT = 100

# --------------------------- Data integrity utils. ---------------------------


class MissingMetrics(NamedTuple):
    objective: dict[str, set[TArmTrial]]
    outcome_constraints: dict[str, set[TArmTrial]]
    tracking_metrics: dict[str, set[TArmTrial]]


def get_missing_metrics(
    data: Data, optimization_config: OptimizationConfig
) -> MissingMetrics:
    """Return all arm_name, trial_index pairs, for which some of the
    observations of optimization config metrics are missing.

    Args:
        data: Data to search.
        optimization_config: provides metric_names to search for.

    Returns:
        A NamedTuple(missing_objective, Dict[str, missing_outcome_constraint])
    """
    objective = optimization_config.objective
    if isinstance(objective, MultiObjective):
        objective_metric_names = [m.name for m in objective.metrics]
    else:
        objective_metric_names = [optimization_config.objective.metric.name]

    outcome_constraints_metric_names = [
        outcome_constraint.metric.name
        for outcome_constraint in optimization_config.outcome_constraints
    ]
    missing_objectives = {
        objective_metric_name: _get_missing_arm_trial_pairs(data, objective_metric_name)
        for objective_metric_name in objective_metric_names
    }
    missing_outcome_constraints = get_missing_metrics_by_name(
        data, outcome_constraints_metric_names
    )
    all_metric_names = set(data.df["metric_name"])
    optimization_config_metric_names = set(missing_objectives.keys()).union(
        outcome_constraints_metric_names
    )
    missing_tracking_metric_names = all_metric_names.difference(
        optimization_config_metric_names
    )
    missing_tracking_metrics = get_missing_metrics_by_name(
        data=data, metric_names=missing_tracking_metric_names
    )
    return MissingMetrics(
        objective={k: v for k, v in missing_objectives.items() if len(v) > 0},
        outcome_constraints={
            k: v for k, v in missing_outcome_constraints.items() if len(v) > 0
        },
        tracking_metrics={
            k: v for k, v in missing_tracking_metrics.items() if len(v) > 0
        },
    )


def get_missing_metrics_by_name(
    data: Data, metric_names: Iterable[str]
) -> dict[str, set[TArmTrial]]:
    """Return all arm_name, trial_index pairs missing some observations of
    specified metrics.

    Args:
        data: Data to search.
        metric_names: list of metrics to search for.

    Returns:
        A Dict[str, missing_metrics], one entry for each metric_name.
    """
    missing_metrics = {
        metric_name: _get_missing_arm_trial_pairs(data=data, metric_name=metric_name)
        for metric_name in metric_names
    }
    return missing_metrics


def _get_missing_arm_trial_pairs(data: Data, metric_name: str) -> set[TArmTrial]:
    """Return arm_name and trial_index pairs missing a specified metric."""
    metric_df = data.df[data.df.metric_name == metric_name]
    present_metric_df = metric_df[metric_df["mean"].notnull()]
    arm_trial_pairs = set(zip(data.df["arm_name"], data.df["trial_index"]))
    arm_trial_pairs_with_metric = set(
        zip(present_metric_df["arm_name"], present_metric_df["trial_index"])
    )
    missing_arm_trial_pairs = arm_trial_pairs.difference(arm_trial_pairs_with_metric)
    return missing_arm_trial_pairs


# ------------------- Utils shared by Client and BatchClient--------------------
def _maybe_update_trial_status_to_complete(
    experiment: Experiment,
    trial_index: int,
) -> None:
    """Check if a trial has all relevant metrics and mark it as completed.
    If the trial has all relevant metrics, mark it as completed.
    If the trial is missing metrics, mark it as failed.

    Args:
        experiment: The experiment to check.
        trial_index: The index of the trial to check.
    """
    if experiment.optimization_config is None:
        raise UserInputError(
            "Cannot attempt to mark a trial as failed without an optimization"
            " config on the expeirment"
        )
    optimization_config = experiment.optimization_config
    trial_data = experiment.lookup_data(trial_indices=[trial_index])
    missing_metrics = {*optimization_config.metrics.keys()} - {*trial_data.metric_names}

    # If all necessary metrics are present mark the trial as COMPLETED
    if len(missing_metrics) == 0:
        experiment.trials[trial_index].mark_completed()
        return

    # If any metrics are missing mark the trial as FAILED
    logger.warning(
        f"Trial {trial_index} marked completed but metrics "
        f"{missing_metrics} are missing, marking trial FAILED."
    )
    experiment.trials[trial_index].mark_failed(
        reason=f"{missing_metrics} are missing, marking trial FAILED."
    )


# -------------------- Experiment result extraction utils. ---------------------


def best_feasible_objective(
    optimization_config: OptimizationConfig,
    values: dict[str, npt.NDArray],
) -> npt.NDArray:
    """Compute the best feasible objective value found by each iteration.

    Args:
        optimization_config: Optimization config.
        values: Dictionary from metric name to array of value at each
            iteration. If optimization config contains outcome constraints, values
            for them must be present in `values`.

    Returns: Array of cumulative best feasible value.
    """
    # Get objective at each iteration
    objective = optimization_config.objective
    f = values[objective.metric.name]
    # Set infeasible points to have infinitely bad values
    infeas_val = np.inf if objective.minimize else -np.inf
    for oc in optimization_config.outcome_constraints:
        if oc.relative:
            raise ValueError(
                "Benchmark aggregation does not support relative constraints"
            )
        g = values[oc.metric.name]
        feas = g <= oc.bound if oc.op == ComparisonOp.LEQ else g >= oc.bound
        f[~feas] = infeas_val

    # Get cumulative best
    minimize = objective.minimize
    accumulate = np.minimum.accumulate if minimize else np.maximum.accumulate
    return accumulate(f)


def _extract_generator_runs(trial: BaseTrial) -> list[GeneratorRun]:
    if isinstance(trial, BatchTrial):
        return trial.generator_runs
    if isinstance(trial, Trial):
        return [none_throws(trial.generator_run)]
    raise ValueError("Unexpected trial type")


def get_model_trace_of_times(
    experiment: Experiment,
) -> tuple[list[float | None], list[float | None]]:
    """
    Get time spent fitting the model and generating candidates during each trial.
    Not cumulative.

    Returns:
        List of fit times, list of gen times.
    """
    generator_runs = [
        gr
        for trial in experiment.trials.values()
        for gr in _extract_generator_runs(trial=trial)
    ]
    fit_times = [gr.fit_time for gr in generator_runs]
    gen_times = [gr.gen_time for gr in generator_runs]
    return fit_times, gen_times


def get_model_times(experiment: Experiment) -> tuple[float, float]:
    """Get total times spent fitting the model and generating candidates in the
    course of the experiment.
    """
    fit_times, gen_times = get_model_trace_of_times(experiment)
    fit_time = sum(t for t in fit_times if t is not None)
    gen_time = sum(t for t in gen_times if t is not None)
    return fit_time, gen_time


def is_bandit_experiment(generation_strategy_name: str) -> bool:
    """
    Determine if an experiment is a bandit experiment based on the generation
    strategy name.

    A bandit experiment is identified by having a generation strategy with the name
    "FACTORIAL + EMPIRICAL_BAYES_THOMPSON_SAMPLING".

    Args:
        generation_strategy_name: Name of the generation_strategy of the experiment
            that needs to be checked.

    Returns:
        True if the generation strategy indicates this is a bandit experiment,
        False otherwise.
    """
    return (
        generation_strategy_name
        == Keys.FACTORIAL_PLUS_EMPIRICAL_BAYES_THOMPSON_SAMPLING
    )


# -------------------- Pending observations extraction utils. ---------------------


def extract_pending_observations(
    experiment: Experiment,
    include_out_of_design_points: bool = False,
) -> dict[str, list[ObservationFeatures]] | None:
    """Computes a list of pending observation features (corresponding to:
    - arms that have been generated and run in the course of the experiment,
    but have not been completed with data,
    - arms that have been abandoned or belong to abandoned trials).

    This function dispatches to:
    - ``get_pending_observation_features`` if experiment is using
    ``BatchTrial``-s or has fewer than 100 trials,
    - ``get_pending_observation_features_based_on_trial_status`` if
    experiment is using  ``Trial``-s and has more than 100 trials.

    ``get_pending_observation_features_based_on_trial_status`` is a faster
    way to compute pending observations, but it is not guaranteed to be
    accurate for ``BatchTrial`` settings and makes assumptions, e.g.
    arms in ``COMPLETED`` trial never being pending. See docstring of
    that function for more details.

    NOTE: Pending observation features are passed to the model to
    instruct it to not generate the same points again.
    """
    if len(experiment.trials) >= MANY_TRIALS_IN_EXPERIMENT and all(
        isinstance(t, Trial) for t in experiment.trials.values()
    ):
        return get_pending_observation_features_based_on_trial_status(
            experiment=experiment,
            include_out_of_design_points=include_out_of_design_points,
        )

    return get_pending_observation_features(
        experiment=experiment, include_out_of_design_points=include_out_of_design_points
    )


def get_pending_observation_features(
    experiment: Experiment,
    *,
    include_out_of_design_points: bool = False,
) -> dict[str, list[ObservationFeatures]] | None:
    """Computes a list of pending observation features (corresponding to:
    - arms that have been generated in the course of the experiment,
    but have not been completed with data,
    - arms that have been abandoned or belong to abandoned trials).

    NOTE: Pending observation features are passed to the model to
    instruct it to not generate the same points again.

    Args:
        experiment: Experiment, pending features on which we seek to compute.
        include_out_of_design_points: By default, this function will not include
            "out of design" points (those that are not in the search space) among
            the pending points. This is because pending points are generally used to
            help the model avoid re-suggesting the same points again. For points
            outside of the search space, this will not happen, so they typically do
            not need to be included. However, if the user wants to include them,
            they can be included by setting this flag to ``True``.

    Returns:
        An optional mapping from metric names to a list of observation features,
        pending for that metric (i.e. do not have evaluation data for that metric).
        If there are no pending features for any of the metrics, return is None.
    """

    def _is_in_design(arm: Arm) -> bool:
        return experiment.search_space.check_membership(parameterization=arm.parameters)

    pending_features = {metric_name: [] for metric_name in experiment.metrics}

    def create_observation_feature(
        arm: Arm,
        trial_index: int,
        trial: BaseTrial,
    ) -> ObservationFeatures | None:
        if not include_out_of_design_points and not _is_in_design(arm=arm):
            return None
        return ObservationFeatures.from_arm(
            arm=arm,
            trial_index=trial_index,
            metadata=trial._get_candidate_metadata(arm_name=arm.name),
        )

    # Note that this assumes that if a metric appears in fetched data, the trial is
    # not pending for the metric. Where only the most recent data matters, this will
    # work, but may need to add logic to check previously added data objects, too.
    for trial_index, trial in experiment.trials.items():
        if trial.status.is_deployed:
            metric_names_in_data = set(trial.lookup_data().df.metric_name.values)
        else:
            metric_names_in_data = set()

        for metric_name in experiment.metrics:
            if metric_name not in pending_features:
                pending_features[metric_name] = []

            if (
                trial.status.is_abandoned
                or trial.status.is_candidate
                or (
                    trial.status.is_deployed
                    and metric_name not in metric_names_in_data
                    and trial.arms is not None
                )
            ):
                for arm in trial.arms:
                    if feature := create_observation_feature(
                        arm=arm,
                        trial_index=trial_index,
                        trial=trial,
                    ):
                        pending_features[metric_name].append(feature)

            # Also add abandoned arms as pending for all metrics.
            if isinstance(trial, BatchTrial):
                for arm in trial.abandoned_arms:
                    if feature := create_observation_feature(
                        arm=arm,
                        trial_index=trial_index,
                        trial=trial,
                    ):
                        pending_features[metric_name].append(feature)

    return pending_features if any(x for x in pending_features.values()) else None


# TODO: allow user to pass search space which overrides that on the experiment
# (to use for the `include_out_of_design_points` check)
def get_pending_observation_features_based_on_trial_status(
    experiment: Experiment,
    include_out_of_design_points: bool = False,
) -> dict[str, list[ObservationFeatures]] | None:
    """A faster analogue of ``get_pending_observation_features`` that makes
    assumptions about trials in experiment in order to speed up extraction
    of pending points.

    Assumptions:

    * All arms in all trials in ``CANDIDATE``, ``STAGED``, ``RUNNING`` and ``ABANDONED``
      statuses are to be considered pending for all outcomes.
    * All arms in all trials in other statuses are to be considered not pending for
      all outcomes.

    This entails:

    * No actual data-fetching for trials to determine whether arms in them are pending
      for specific outcomes.
    * Even if data is present for some outcomes in ``RUNNING`` trials, their arms will
      still be considered pending for those outcomes.

    NOTE: This function should not be used to extract pending features in field
    experiments, where arms in running trials should not be considered pending if
    there is data for those arms.

    Args:
        experiment: Experiment, pending features on which we seek to compute.

    Returns:
        An optional mapping from metric names to a list of observation features,
        pending for that metric (i.e. do not have evaluation data for that metric).
        If there are no pending features for any of the metrics, return is None.
    """
    pending_features_list = []
    for status in [
        TrialStatus.CANDIDATE,
        TrialStatus.STAGED,
        TrialStatus.RUNNING,
        TrialStatus.ABANDONED,
    ]:
        for trial in experiment.trials_by_status[status]:
            for arm in trial.arms:
                if (
                    include_out_of_design_points
                    or experiment.search_space.check_membership(arm.parameters)
                ):
                    pending_features_list.append(
                        ObservationFeatures.from_arm(
                            arm=arm,
                            trial_index=trial.index,
                            metadata=trial._get_candidate_metadata(arm_name=arm.name),
                        )
                    )
    pending_features = {
        # Using deepcopy here to avoid issues due to in-place transforms.
        metric_name: deepcopy(pending_features_list)
        for metric_name in experiment.metrics
    }
    return pending_features if pending_features_list else None


def extend_pending_observations(
    experiment: Experiment,
    pending_observations: dict[str, list[ObservationFeatures]],
    generator_run: GeneratorRun,
) -> None:
    """Extend given pending observations dict (from metric name to observations
    that are pending for that metric), with arms in a given generator run.

    Note: This function performs this operation in-place for performance reasons.
    It is only used within the ``GenerationStrategy`` class, and is not intended
    for wide re-use. Please use caution when re-using this function.

    Args:
        experiment: Experiment, for which the generation strategy is producing
            ``GeneratorRun``s.
        pending_observations: Dict from metric name to pending observations for
            that metric, used to avoid resuggesting arms that will be explored soon.
        generator_run: ``GeneratorRun`` currently produced by the
            ``GenerationStrategy`` to add to the pending points.

    """
    for m in experiment.metrics:
        if m not in pending_observations:
            pending_observations[m] = []
        pending_observations[m].extend(
            ObservationFeatures.from_arm(a) for a in generator_run.arms
        )
    return


# -------------------- Get target trial utils. ---------------------


def get_target_trial_index(experiment: Experiment) -> int | None:
    """Get the index of the target trial in the ``Experiment``.

    Find the target trial, among the trials with data for status quo arm, giving
    priority in the following order:
        1. a running long-run trial. Note if there is a running long-run trial on the
            experiment without data, or if there is no data on the experiment, then
            this will return None.
        2. Most recent trial expecting data with running trials be considered the most
            recent.

    In the event of any ties, the tie breaking order is:
        a. longest running trial by duration
        b. trial with most arms
        c. arbitrary selection

    Args:
        experiment: The experiment associated with this ``GenerationStrategy``.

    Returns:
        The index of the target trial in the ``Experiment``.
    """
    # TODO: @mgarrard improve logic to include trial_obsolete_threshold that
    # takes into account the age of the trial, and consider more heavily weighting
    # long run trials.
    df = experiment.lookup_data().df
    status_quo = experiment.status_quo
    if df.empty or status_quo is None:
        return None
    # Filter to only trials with data for status quo arm.
    df = df[df["arm_name"] == status_quo.name]
    trial_indices_with_data = set(df.trial_index.unique())
    # only consider running trials with data
    running_trials = [
        trial
        for trial in experiment.trials_by_status[TrialStatus.RUNNING]
        if trial.index in trial_indices_with_data
    ]
    sorted_running_trials = _sort_trials(trials=running_trials, trials_are_running=True)
    # Priority 1: Any running long-run trial
    has_running_long_run_trial = any(
        trial.trial_type == Keys.LONG_RUN
        for trial in experiment.trials_by_status[TrialStatus.RUNNING]
    )
    if has_running_long_run_trial:
        # This returns a running long-run trial with data or None
        # if there are running long-run trials on the experiment, but
        # no data for that trial
        return next(
            (
                trial.index
                for trial in sorted_running_trials
                if trial.trial_type == Keys.LONG_RUN
            ),
            None,
        )

    # Priority 2: longest running currently running trial with data
    if len(sorted_running_trials) > 0:
        return sorted_running_trials[0].index

    # Priortiy 3: the longest running trial with data, discounting running trials
    # as we handled those above
    non_running_trial_indices_with_data = trial_indices_with_data - {
        t.index for t in running_trials
    }
    non_running_trials_with_data = [
        experiment.trials[i] for i in non_running_trial_indices_with_data
    ]
    sorted_non_running_trials_with_data = _sort_trials(
        trials=non_running_trials_with_data, trials_are_running=False
    )
    if len(sorted_non_running_trials_with_data) > 0:
        return sorted_non_running_trials_with_data[0].index

    return None


def _sort_trials(
    trials: list[BaseTrial],
    trials_are_running: bool,
) -> list[BaseTrial]:
    """Sort a list of trials by (1) duration of trial, (2) number of arms in trial.

    Args:
        trials: The trials to sort.
        trials_are_running: Whether the trials are running or not, used to determine
            the trial duration for sorting

    Returns:
        The sorted trials.
    """
    default_time_run_started = datetime.now()
    twelve_hours_in_secs = 12 * 60 * 60
    sorted_trials_expecting_data = sorted(
        trials,
        key=lambda t: (
            # First sorting criterion: trial duration, if a trial's duration is within
            # 12 hours of another trial, we consider them to be a tie
            int(
                (
                    # if the trial is running, we set end time to now for sorting ease
                    (
                        _time_trial_completed_safe(trial=t).timestamp()
                        if not trials_are_running
                        else default_time_run_started.timestamp()
                    )
                    - _time_trial_started_safe(
                        trial=t, default_time_run_started=default_time_run_started
                    ).timestamp()
                )
                // twelve_hours_in_secs
            ),
            # In event of a tie, we want the trial with the most arms
            +len(t.arms_by_name),
        ),
        reverse=True,
    )
    return sorted_trials_expecting_data


def _time_trial_started_safe(
    trial: BatchTrial, default_time_run_started: datetime
) -> datetime:
    """Not all RUNNING trials have ``time_run_started`` defined.
    This function accepts, but penalizes those trials by using a
    default ``time_run_started``, which moves them to the end of
    the sort because they would be running a very short time.

    Args:
        trial: The trial to check.
        default_time_run_started: The time to use if `time_run_started` is not defined.
            Do not use ``default_time_run_started=datetime.now()`` as it will be
            slightly different for each trial.  Instead set ``val = datetime.now()``
            and then pass ``val`` as the ``default_time_run_started`` argument.
    """
    return (
        trial.time_run_started
        if trial.time_run_started is not None
        else default_time_run_started
    )


def _time_trial_completed_safe(trial: BatchTrial) -> datetime:
    """Not all COMPLETED trials have `time_completed` defined.
    This functions accepts, but penalizes those trials by
    choosing epoch 0 as the completed time,
    which moves them to the end of the sort because
    they would be running a very short time."""
    return (
        trial.time_completed
        if trial.time_completed is not None
        else datetime.fromtimestamp(0)
    )


# -------------------- MapMetric related utils. ---------------------


def has_map_metrics(optimization_config: OptimizationConfig) -> bool:
    """Check if the optimization config has any ``MapMetric``s.

    Args:
        optimization_config: Optimization config.
    """
    metrics = optimization_config.metrics
    # Technically an OptimizationConfig could have zero metrics since a
    # MultiObjective could have 0 objectives
    if len(metrics) == 0:
        return False
    return any(isinstance(metric, MapMetric) for metric in metrics.values())


# -------------------- Context manager and decorator utils. ---------------------


def batch_trial_only(msg: str | None = None) -> Callable[..., Any]:
    """A decorator to verify that the value passed to the `trial`
    argument to `func` is a `BatchTrial`.
    """

    def batch_trial_only_decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(func)
        def _batch_trial_only(*args: Any, **kwargs: Any) -> Any:
            if "trial" not in kwargs:
                raise AxError(
                    f"Expected a keyword argument `trial` to `{func.__name__}`."
                )
            if not isinstance(kwargs["trial"], BatchTrial):
                message = msg or (
                    f"Expected the argument `trial` to `{func.__name__}` "
                    f"to be a `BatchTrial`, but got {type(kwargs['trial'])}."
                )
                raise AxError(message)
            return func(*args, **kwargs)

        return _batch_trial_only

    return batch_trial_only_decorator



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import pandas as pd
from ax.analysis.plotly.parallel_coordinates import (
    _get_parameter_dimension,
    ParallelCoordinatesPlot,
)
from ax.analysis.plotly.utils import select_metric
from ax.exceptions.core import UnsupportedError, UserInputError
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import (
    get_branin_experiment,
    get_experiment_with_multi_objective,
    get_experiment_with_scalarized_objective_and_outcome_constraint,
    get_offline_experiments,
    get_online_experiments,
)
from pyre_extensions import none_throws


class TestParallelCoordinatesPlot(TestCase):
    def test_compute(self) -> None:
        analysis = ParallelCoordinatesPlot("branin")
        experiment = get_branin_experiment(with_completed_trial=True)

        with self.assertRaisesRegex(UserInputError, "requires an Experiment"):
            analysis.compute()

        card = analysis.compute(experiment=experiment)
        self.assertEqual(card.name, "ParallelCoordinatesPlot")
        self.assertEqual(card.title, "Parallel Coordinates for branin")
        self.assertEqual(
            card.subtitle,
            (
                "The parallel coordinates plot displays multi-dimensional "
                "data by representing each parameter as a parallel axis. This "
                "plot helps in assessing how thoroughly the search space has "
                "been explored and in identifying patterns or clusterings associated "
                "with high-performing (good) or low-performing (bad) arms. By "
                "tracing lines across the axes, one can observe correlations and "
                "interactions between parameters, gaining insights into the "
                "relationships that contribute to the success or failure of "
                "different configurations within the experiment."
            ),
        )
        self.assertEqual(
            {*card.df.columns}, {"trial_index", "arm_name", "branin", "x1", "x2"}
        )
        self.assertIsNotNone(card.blob)

        analysis_no_metric = ParallelCoordinatesPlot()
        _ = analysis_no_metric.compute(experiment=experiment)

    def test_select_metric(self) -> None:
        experiment = get_branin_experiment()
        experiment_no_optimization_config = get_branin_experiment(
            has_optimization_config=False
        )
        experiment_multi_objective = get_experiment_with_multi_objective()
        experiment_scalarized_objective = (
            get_experiment_with_scalarized_objective_and_outcome_constraint()
        )

        self.assertEqual(select_metric(experiment=experiment), "branin")

        with self.assertRaisesRegex(ValueError, "OptimizationConfig"):
            select_metric(experiment=experiment_no_optimization_config)

        with self.assertRaisesRegex(UnsupportedError, "MultiObjective"):
            select_metric(experiment=experiment_multi_objective)

        with self.assertRaisesRegex(UnsupportedError, "ScalarizedObjective"):
            select_metric(experiment=experiment_scalarized_objective)

    def test_get_parameter_dimension(self) -> None:
        range_series = pd.Series([0, 1, 2, 3], name="range")
        range_dimension = _get_parameter_dimension(series=range_series)
        self.assertEqual(
            range_dimension,
            {
                "tickvals": None,
                "ticktext": None,
                "label": "range",
                "values": range_series.tolist(),
            },
        )

        choice_series = pd.Series(["foo", "bar", "baz"], name="choice")
        choice_dimension = _get_parameter_dimension(series=choice_series)
        self.assertEqual(
            choice_dimension,
            {
                "tickvals": ["0", "1", "2"],
                "ticktext": ["bar", "baz", "foo"],
                "label": "choice",
                "values": [2, 0, 1],
            },
        )

    def test_online(self) -> None:
        # Test ParallelCoordinatesPlot can be computed for a variety of experiments
        # which resemble those we see in an online setting.

        for experiment in get_online_experiments():
            analysis = ParallelCoordinatesPlot(
                # Select and arbitrary metric from the optimization config
                metric_name=none_throws(
                    experiment.optimization_config
                ).objective.metric_names[0]
            )

            _ = analysis.compute(experiment=experiment)

    def test_offline(self) -> None:
        # Test ParallelCoordinatesPlot can be computed for a variety of experiments
        # which resemble those we see in an offline setting.

        for experiment in get_offline_experiments():
            analysis = ParallelCoordinatesPlot(
                # Select and arbitrary metric from the optimization config
                metric_name=none_throws(
                    experiment.optimization_config
                ).objective.metric_names[0]
            )

            _ = analysis.compute(experiment=experiment)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import numpy as np
from ax.generators.discrete.ashr_utils import (
    Ashr,
    fit_ashr_em,
    GaussianMixture,
    marginal_densities,
    prior_grid,
)
from ax.utils.common.testutils import TestCase


class AshrModelTest(TestCase):
    def test_gaussian_mixture(self) -> None:
        normal_vars_by_class = np.array(
            [[0.0, 1.0, 1.0], [0.0, 1.0, 0.25], [0.0, 2.25, 1.0], [0.0, 4.0, 4.0]]
        )

        gm = GaussianMixture(
            normal_means_by_class=np.zeros((4, 3)),
            normal_vars_by_class=normal_vars_by_class,
            weights=np.ones((4, 3)) / 3.0,
        )

        variances = normal_vars_by_class.sum(axis=1) / 3.0
        self.assertTrue(np.allclose(gm.means, np.zeros(4)))
        self.assertTrue(np.allclose(gm.vars, variances))
        self.assertTrue(
            np.allclose(gm.tail_probabilities(left_tail=True), np.ones(4) / 3)
        )
        self.assertTrue(
            np.allclose(gm.tail_probabilities(left_tail=False), np.ones(4) / 3)
        )

    def test_prior_grid(self) -> None:
        prior_stds = prior_grid(
            Y=np.array([1, 2]), Yvar=np.array([1, 1]), grid_param=16.0
        )
        self.assertTrue((prior_stds == np.array([0.0, 0.1, 0.4, 1.6, 6.4])).all())

    def test_marginal_densities(self) -> None:
        ll = marginal_densities(
            Y=np.zeros(2), Yvar=np.array([1, 2]), prior_vars=np.array([0, 1])
        )
        self.assertTrue(
            (
                ll
                == np.array(
                    [
                        [1 / np.sqrt(2 * np.pi), 1 / np.sqrt(2 * 2 * np.pi)],
                        [1 / np.sqrt(2 * 2 * np.pi), 1 / np.sqrt(3 * 2 * np.pi)],
                    ]
                )
            ).all()
        )

    def test_fit_ashr_em(self) -> None:
        results = fit_ashr_em(
            ll=np.array([[1, 2], [0, 3]]), lambdas=np.ones(2), threshold=1.0, nsteps=1
        )
        self.assertTrue(np.allclose(results["pi"], np.array([1 / 6, 5 / 6])))
        self.assertTrue(
            np.allclose(
                results["weights"], np.array([[1.0 / 11, 10.0 / 11], [0.0, 1.0]])
            )
        )
        self.assertTrue(np.allclose(results["lfdr"], np.array([1.0 / 11, 0.0])))

    def test_ashr_posterior(self) -> None:
        a = Ashr(Y=np.ones(2), Yvar=np.ones(2), prior_vars=np.array([0, 1]))
        gm = a.posterior(w=np.ones((2, 2)))
        self.assertTrue(
            (gm.normal_means_by_class == np.array([[0, 0.5], [0, 0.5]])).all()
        )
        self.assertTrue(
            (gm.normal_vars_by_class == np.array([[0, 0.5], [0, 0.5]])).all()
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from typing import Any

from ax.exceptions.core import UnsupportedError
from botorch.acquisition.acquisition import AcquisitionFunction
from botorch.acquisition.knowledge_gradient import qKnowledgeGradient
from botorch.acquisition.multioutput_acquisition import MultiOutputAcquisitionFunction

from .optimizer_defaults import (
    BATCH_LIMIT,
    INIT_BATCH_LIMIT,
    MAX_OPT_AGG_SIZE,
    NUM_RESTARTS,
    RAW_SAMPLES,
)


def optimizer_argparse(
    acqf: AcquisitionFunction,
    *,
    optimizer: str,
    optimizer_options: dict[str, Any] | None = None,
) -> dict[str, Any]:
    """Extract the kwargs to be passed to a BoTorch optimizer.

    Args:
        acqf: The acquisition function being optimized.
        optimizer: The optimizer to parse args for. Typically chosen by
            `Acquisition.optimize`. Must be one of:
            - "optimize_acqf",
            - "optimize_acqf_discrete_local_search",
            - "optimize_acqf_discrete",
            - "optimize_acqf_homotopy",
            - "optimize_acqf_mixed",
            - "optimize_acqf_mixed_alternating".
        optimizer_options: An optional dictionary of optimizer options (some of
            these under an `options` dictionary); default values will be used
            where not specified. See the docstrings in
            `botorch/optim/optimize.py` for supported options.
            Example:
                >>> optimizer_options = {
                >>>     "num_restarts": 20,
                >>>     "options": {
                >>>         "maxiter": 200,
                >>>         "batch_limit": 5,
                >>>     },
                >>>     "retry_on_optimization_warning": False,
                >>> }
    """
    supported_optimizers = [
        "optimize_acqf",
        "optimize_acqf_discrete_local_search",
        "optimize_acqf_discrete",
        "optimize_acqf_homotopy",
        "optimize_acqf_mixed",
        "optimize_acqf_mixed_alternating",
        "optimize_with_nsgaii",
    ]
    if optimizer not in supported_optimizers:
        raise ValueError(
            f"optimizer=`{optimizer}` is not supported. Accepted options are "
            f"{supported_optimizers}"
        )
    if (optimizer != "optimize_acqf") and isinstance(acqf, qKnowledgeGradient):
        raise RuntimeError(
            "Ax is attempting to use a discrete or mixed optimizer, "
            f"`{optimizer}`, but this is not compatible with "
            "`qKnowledgeGradient` or its subclasses. To address this, please "
            "either use a different acquisition class or make parameters "
            "continuous using the transform "
            "`ax.adapter.registry.Cont_X_trans`."
        )
    elif optimizer == "optimize_with_nsgaii" and not isinstance(
        acqf, MultiOutputAcquisitionFunction
    ):
        raise RuntimeError(
            "Ax is attempting to use a NSGA-II optimizer, "
            f"`{optimizer}`, but this is not compatible with "
            "single-objective acquisition functions. To address this, please "
            "use MultiAcquisitionFunction or use a different optimizer."
        )
    provided_options = optimizer_options if optimizer_options is not None else {}

    # Construct arguments from options that are not `provided_options`.
    if optimizer in ("optimize_acqf_discrete", "optimize_with_nsgaii"):
        # `optimize_acqf_discrete` only accepts 'choices', 'max_batch_size', 'unique'.
        options = {}
    else:
        options = {
            "num_restarts": NUM_RESTARTS,
            "raw_samples": RAW_SAMPLES,
        }

    if optimizer in [
        "optimize_acqf",
        "optimize_acqf_homotopy",
        "optimize_acqf_mixed",
        "optimize_acqf_mixed_alternating",
    ]:
        options["options"] = {
            "init_batch_limit": INIT_BATCH_LIMIT,
            "batch_limit": BATCH_LIMIT,
            **provided_options.get("options", {}),
        }

        if optimizer == "optimize_acqf":
            options["options"]["max_optimization_problem_aggregation_size"] = (
                MAX_OPT_AGG_SIZE
            )
    # Error out if options are specified for an optimizer that does not support the arg.
    elif "options" in provided_options:
        raise UnsupportedError(
            f"`{optimizer=}` does not support the `options` argument."
        )

    if optimizer == "optimize_acqf":
        options["sequential"] = True

    options.update(**{k: v for k, v in provided_options.items() if k != "options"})
    return options



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from collections.abc import Mapping
from dataclasses import dataclass

import pandas as pd

from ax.utils.testing.backend_simulator import BackendSimulator


@dataclass(kw_only=True, frozen=True)
class BenchmarkTrialMetadata:
    """
    Data pertaining to one trial evaluation.

    Args:
        df: A dict mapping each metric name to a Pandas DataFrame with columns
            ["metric_name", "arm_name", "mean", "sem", and "step"]. The "sem" is
            always present in this df even if noise levels are unobserved;
            ``BenchmarkMetric`` and ``BenchmarkMapMetric`` hide that data if it
            should not be observed, and ``BenchmarkMapMetric``s drop data from
            time periods that that are not observed based on the (simulated)
            trial progression.
        backend_simulator: Optionally, the backend simulator that is tracking
            the trial's status.
    """

    dfs: Mapping[str, pd.DataFrame]
    backend_simulator: BackendSimulator | None = None



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

import traceback
from logging import Logger
from typing import Protocol, Sequence

import pandas as pd
from ax.adapter.base import Adapter
from ax.analysis.analysis_card import (
    AnalysisCard,
    AnalysisCardBase,
    AnalysisCardGroup,
    ErrorAnalysisCard,
)
from ax.core.experiment import Experiment
from ax.generation_strategy.generation_strategy import GenerationStrategy
from ax.utils.common.logger import get_logger
from ax.utils.common.result import Err, ExceptionE, Ok, Result
from IPython.display import display

logger: Logger = get_logger(__name__)


class Analysis(Protocol):
    """
    An Analysis is a class that given either and Experiment, a GenerationStrategy, or
    both can compute some data intended for end-user consumption. The data is returned
    to the user in the form of an AnalysisCard which contains the raw data, a blob (the
    data processed for end-user consumption), and miscellaneous metadata that can be
    useful for rendering the card or a collection of cards.

    The AnalysisCard is a thin wrapper around the raw data and the processed blob;
    Analyses impose structure on their blob should subclass Analysis. See
    PlotlyAnalysis for an example which produces cards where the blob is always a
    Plotly Figure object.

    A good pattern to follow when implementing your own Analyses is to configure
    "settings" (like which parameter or metrics to operate on, or whether to use
    observed or modeled effects) in your Analyses' __init__ methods, then to consume
    these settings in the compute method.
    """

    def compute(
        self,
        experiment: Experiment | None = None,
        generation_strategy: GenerationStrategy | None = None,
        adapter: Adapter | None = None,
    ) -> AnalysisCardBase:
        # Note: when implementing compute always prefer experiment.lookup_data() to
        # experiment.fetch_data() to avoid unintential data fetching within the report
        # generation.
        ...

    def compute_result(
        self,
        experiment: Experiment | None = None,
        generation_strategy: GenerationStrategy | None = None,
        adapter: Adapter | None = None,
    ) -> Result[AnalysisCardBase, AnalysisE]:
        """
        Utility method to compute an AnalysisCard as a Result. This can be useful for
        computing many Analyses at once and handling Exceptions later.
        """

        try:
            card = self.compute(
                experiment=experiment,
                generation_strategy=generation_strategy,
                adapter=adapter,
            )
            return Ok(value=card)

        except Exception as e:
            logger.error(f"Failed to compute {self.__class__.__name__}")
            logger.error(traceback.format_exc())

            return Err(
                value=AnalysisE(
                    message=f"Failed to compute {self.__class__.__name__}",
                    exception=e,
                    analysis=self,
                )
            )

    def compute_or_error_card(
        self,
        experiment: Experiment | None = None,
        generation_strategy: GenerationStrategy | None = None,
        adapter: Adapter | None = None,
    ) -> AnalysisCardBase:
        """
        Utility method to compute an AnalysisCard or an ErrorAnalysisCard if an
        exception is raised.
        """
        return self.compute_result(
            experiment=experiment,
            generation_strategy=generation_strategy,
            adapter=adapter,
        ).unwrap_or_else(error_card_from_analysis_e)

    def _create_analysis_card(
        self,
        title: str,
        subtitle: str,
        df: pd.DataFrame,
    ) -> AnalysisCard:
        """
        Make an AnalysisCard from this Analysis using provided fields and
        details about the Analysis class.
        """
        return AnalysisCard(
            name=self.__class__.__name__,
            title=title,
            subtitle=subtitle,
            df=df,
            blob=df.to_json(),
        )

    def _create_analysis_card_group(
        self,
        title: str,
        subtitle: str | None,
        children: Sequence[AnalysisCardBase],
    ) -> AnalysisCardGroup:
        """
        Make an AnalysisCardGroup from this Analysis using provided fields and
        details about the Analysis class.
        """
        return AnalysisCardGroup(
            name=self.__class__.__name__,
            title=title,
            subtitle=subtitle if subtitle is not None else "",
            children=children,
        )

    def _create_analysis_card_group_or_card(
        self,
        title: str,
        subtitle: str,
        children: Sequence[AnalysisCardBase],
    ) -> AnalysisCardBase:
        """
        Make an AnalysisCardGroup from this Analysis using provided fields and
        details about the Analysis class. If there is only one child, return the child
        directly instead of wrapping it in a group.
        """
        if len(children) == 1:
            return children[0]

        return self._create_analysis_card_group(
            title=title,
            subtitle=subtitle,
            children=children,
        )


class AnalysisE(ExceptionE):
    analysis: Analysis

    def __init__(
        self,
        message: str,
        exception: Exception,
        analysis: Analysis,
    ) -> None:
        super().__init__(message, exception)
        self.analysis = analysis


def display_cards(
    cards: Sequence[AnalysisCardBase],
) -> None:
    """
    Helper method for displaying a sequence of AnalysisCards (as is returned by adhoc
    compute_ methods and by Client.compute_analyses).
    """
    for card in cards:
        display(card)


def error_card_from_analysis_e(
    analysis_e: AnalysisE,
) -> ErrorAnalysisCard:
    analysis_name = analysis_e.analysis.__class__.__name__
    exception_name = analysis_e.exception.__class__.__name__

    return ErrorAnalysisCard(
        name=analysis_name,
        title=f"{analysis_name} Error",
        subtitle=f"{exception_name} encountered while computing {analysis_name}.",
        df=pd.DataFrame(),
        blob=analysis_e.tb_str() or "",
    )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import pandas as pd
from ax.analysis.plotly.utils import get_arm_tooltip, trial_index_to_color
from ax.core.trial_status import TrialStatus
from ax.utils.common.testutils import TestCase


class TestUtils(TestCase):
    def test_trial_index_to_color(self) -> None:
        trials_list = [0, 1, 11]
        test_df = pd.DataFrame(
            {
                "trial_index": [0, 1, 11, 15],  # Trial 15 is a candidate trial
                "trial_status": [
                    TrialStatus.COMPLETED.name,
                    TrialStatus.COMPLETED.name,
                    TrialStatus.COMPLETED.name,
                    TrialStatus.CANDIDATE.name,
                ],
            }
        )
        # Test last completed trial is Botorch Blue
        botorch_blue_no_transparency = trial_index_to_color(
            trial_df=test_df.iloc[[2]],
            trials_list=trials_list,
            trial_index=11,
            transparent=False,
        )
        expected_color = (
            "rgba(76, 110, 243, 1)"  # RGB for Botorch Blue with no transparency
        )
        self.assertEqual(botorch_blue_no_transparency, expected_color)

        # Test last completed trial is Botorch Blue with transparency
        botorch_blue_with_transparency = trial_index_to_color(
            trial_df=test_df.iloc[[2]],
            trials_list=trials_list,
            trial_index=11,
            transparent=True,
        )
        expected_color_with_transparency = (
            "rgba(76, 110, 243, 0.5)"  # RGB for #4C6EF3 with transparency
        )
        self.assertEqual(
            botorch_blue_with_transparency, expected_color_with_transparency
        )

        # Test candidate trial is LIGHT_AX_BLUE with transparency
        candidate_light_ax_blue_with_transparency = trial_index_to_color(
            trial_df=test_df.iloc[[3]],
            trials_list=trials_list,
            trial_index=15,
            transparent=True,
        )
        expected_color_for_candidate = (
            "rgba(173, 192, 253, 0.5)"  # RGB for LIGHT_AX_BLUE with transparency
        )
        self.assertEqual(
            candidate_light_ax_blue_with_transparency, expected_color_for_candidate
        )

    def test_get_arm_tooltip(self) -> None:
        row = pd.Series(
            {
                "trial_index": 5,
                "trial_status": "COMPLETED",
                "arm_name": "5_0",
                "foo_mean": 0.450508,
                "bar_mean": -0.222466,
                "foo_sem": 0.15232,
                "bar_sem": 0.638425,
                "generation_node": "MBM",
                "p_feasible_mean": 0.1,
            }
        )

        foo_tooltip = get_arm_tooltip(row=row, metric_names=["foo"])
        self.assertEqual(
            foo_tooltip,
            "Trial: 5<br />Arm: 5_0<br />Status: COMPLETED<br />"
            "Generation Node: MBM<br />foo: 0.45051±0.29855<br />",
        )

        bar_tooltip = get_arm_tooltip(row=row, metric_names=["bar"])
        self.assertEqual(
            bar_tooltip,
            "Trial: 5<br />Arm: 5_0<br />Status: COMPLETED<br />"
            "Generation Node: MBM<br />bar: -0.22247±1.25131<br />",
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from unittest.mock import patch

from ax.utils.common.testutils import TestCase


class TestWithDbSetingsBaseMinimalDependencies(TestCase):
    @patch.dict("sys.modules", {"sqlalchemy": None})
    def test_with_db_settings_base_no_sql_alchemy(self) -> None:
        from ax.service.utils.with_db_settings_base import WithDBSettingsBase  # noqa



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from logging import Logger
from math import isnan
from typing import Any, SupportsFloat, TYPE_CHECKING

from ax.adapter.data_utils import ExperimentData
from ax.adapter.transforms.base import Transform
from ax.adapter.transforms.metadata_to_parameter import MetadataToParameterMixin
from ax.core import ParameterType
from ax.core.observation import Observation
from ax.core.parameter import RangeParameter
from ax.core.search_space import SearchSpace
from ax.generators.types import TConfig
from ax.utils.common.logger import get_logger
from pyre_extensions import assert_is_instance, none_throws

if TYPE_CHECKING:
    # import as module to make sphinx-autodoc-typehints happy
    from ax import adapter as adapter_module  # noqa F401


logger: Logger = get_logger(__name__)


class MetadataToFloat(MetadataToParameterMixin, Transform):
    """
    This transform converts metadata from observation features into range (float)
    parameters for a search space.

    It allows the user to specify the `config` with `parameters` as the key, where
    each entry maps a metadata key to a dictionary of keyword arguments for the
    corresponding RangeParameter constructor. NOTE: log and logit-scale options
    are not supported.

    Transform is done in-place.
    """

    requires_data_for_initialization: bool = True

    DEFAULT_IS_FIDELITY: bool = False
    ENFORCE_BOUNDS: bool = False

    def __init__(
        self,
        search_space: SearchSpace | None = None,
        observations: list[Observation] | None = None,
        experiment_data: ExperimentData | None = None,
        adapter: adapter_module.base.Adapter | None = None,
        config: TConfig | None = None,
    ) -> None:
        Transform.__init__(
            self,
            search_space=search_space,
            observations=observations,
            experiment_data=experiment_data,
            adapter=adapter,
            config=config,
        )
        config = config or {}
        self.parameters: dict[str, dict[str, Any]] = assert_is_instance(
            config.get("parameters", {}), dict
        )

        self._parameter_list: list[RangeParameter] = []
        for name in self.parameters:
            values: set[float] = self._get_values_for_parameter(
                name=name, observations=observations, experiment_data=experiment_data
            )

            if len(values) == 0:
                logger.debug(
                    f"Did not encounter any non-NaN values for "
                    f"metadata key '{name}'. Not adding to parameters."
                )
                continue
            if len(values) == 1:
                (value,) = values
                logger.debug(
                    f"Encountered only a single unique value {value:.1f} in "
                    f"metadata key '{name}'. Not adding to parameters."
                )
                continue

            lower: float = self.parameters[name].get("lower", min(values))
            upper: float = self.parameters[name].get("upper", max(values))

            digits = self.parameters[name].get("digits")
            is_fidelity = self.parameters[name].get(
                "is_fidelity", self.DEFAULT_IS_FIDELITY
            )

            target_value = self.parameters[name].get("target_value")

            parameter = RangeParameter(
                name=name,
                parameter_type=ParameterType.FLOAT,
                lower=lower,
                upper=upper,
                digits=digits,
                is_fidelity=is_fidelity,
                target_value=target_value,
            )
            self._parameter_list.append(parameter)

    def _get_values_for_parameter(
        self,
        name: str,
        observations: list[Observation] | None,
        experiment_data: ExperimentData | None,
    ) -> set[float]:
        if experiment_data is not None:
            all_metadata = experiment_data.arm_data["metadata"]
            return all_metadata.str.get(name).dropna().astype(float).tolist()

        values: set[float] = set()
        for obs in none_throws(observations):
            obsf_metadata = none_throws(obs.features.metadata)
            value = float(assert_is_instance(obsf_metadata[name], SupportsFloat))
            if not isnan(value):
                values.add(value)
        return values



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


from collections.abc import Mapping, Sequence

import numpy as np
from ax.adapter.adapter_utils import (
    extract_parameter_constraints,
    extract_search_space_digest,
    get_fixed_features,
    parse_observation_features,
    transform_callback,
)
from ax.adapter.base import Adapter, DataLoaderConfig, GenResults
from ax.adapter.data_utils import ExperimentData
from ax.adapter.transforms.base import Transform
from ax.core.data import Data
from ax.core.experiment import Experiment
from ax.core.observation import ObservationFeatures
from ax.core.optimization_config import OptimizationConfig
from ax.core.search_space import SearchSpace
from ax.generators.random.base import RandomGenerator
from ax.generators.types import TConfig


class RandomAdapter(Adapter):
    """An adapter for using purely random ``RandomGenerator``s.
    Data and optimization configs are not required.

    Please refer to base ``Adapter`` class for documentation of constructor arguments.
    """

    def __init__(
        self,
        *,
        experiment: Experiment,
        generator: RandomGenerator,
        search_space: SearchSpace | None = None,
        data: Data | None = None,
        transforms: Sequence[type[Transform]] | None = None,
        transform_configs: Mapping[str, TConfig] | None = None,
        optimization_config: OptimizationConfig | None = None,
        fit_tracking_metrics: bool = True,
        fit_on_init: bool = True,
        data_loader_config: DataLoaderConfig | None = None,
        fit_out_of_design: bool | None = None,
        fit_abandoned: bool | None = None,
    ) -> None:
        self.parameters: list[str] = []
        super().__init__(
            search_space=search_space,
            generator=generator,
            transforms=transforms,
            experiment=experiment,
            data=data,
            transform_configs=transform_configs,
            optimization_config=optimization_config,
            expand_model_space=False,
            data_loader_config=data_loader_config,
            fit_out_of_design=fit_out_of_design,
            fit_abandoned=fit_abandoned,
            fit_tracking_metrics=fit_tracking_metrics,
            fit_on_init=fit_on_init,
        )
        # Re-assign for more precise typing.
        self.generator: RandomGenerator = generator

    def _fit(
        self,
        search_space: SearchSpace,
        experiment_data: ExperimentData,
    ) -> None:
        """Extracts the list of parameters from the search space."""
        self.parameters = list(search_space.parameters.keys())

    def _gen(
        self,
        n: int,
        search_space: SearchSpace,
        pending_observations: dict[str, list[ObservationFeatures]],
        fixed_features: ObservationFeatures | None,
        optimization_config: OptimizationConfig | None,
        model_gen_options: TConfig | None,
    ) -> GenResults:
        """Generate new candidates according to a search_space."""
        # Extract parameter values
        search_space_digest = extract_search_space_digest(search_space, self.parameters)
        # Get fixed features
        fixed_features_dict = get_fixed_features(fixed_features, self.parameters)
        # Extract param constraints
        linear_constraints = extract_parameter_constraints(
            search_space.parameter_constraints, self.parameters
        )
        # Extract generated points to deduplicate against.
        # Exclude out-of-design arms (which can only be manual arms
        # instead of adapter-generated arms).
        generated_points = None
        if self.generator.deduplicate:
            arms_to_deduplicate = self._experiment.arms_by_signature_for_deduplication
            generated_obs = [
                ObservationFeatures.from_arm(arm=arm)
                for arm in arms_to_deduplicate.values()
                if self._search_space.check_membership(parameterization=arm.parameters)
            ]
            # Transform
            for t in self.transforms.values():
                generated_obs = t.transform_observation_features(generated_obs)
            # Add pending observations -- already transformed.
            generated_obs.extend(
                [obs for obs_list in pending_observations.values() for obs in obs_list]
            )
            if len(generated_obs) > 0:
                # Extract generated points array (n x d).
                generated_points = np.array(
                    [
                        [obs.parameters[p] for p in self.parameters]
                        for obs in generated_obs
                    ]
                )
                # Take unique points only, since there may be duplicates coming
                # from pending observations for different metrics.
                generated_points = np.unique(generated_points, axis=0)

        # Generate the candidates
        X, w = self.generator.gen(
            n=n,
            bounds=search_space_digest.bounds,
            linear_constraints=linear_constraints,
            fixed_features=fixed_features_dict,
            model_gen_options=model_gen_options,
            rounding_func=transform_callback(self.parameters, self.transforms),
            generated_points=generated_points,
        )
        observation_features = parse_observation_features(X, self.parameters)
        return GenResults(
            observation_features=observation_features,
            weights=w.tolist(),
        )

    def _set_status_quo(self, experiment: Experiment) -> None:
        pass



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import pandas as pd
from ax.adapter.base import Adapter
from ax.analysis.analysis import Analysis
from ax.analysis.healthcheck.healthcheck_analysis import (
    create_healthcheck_analysis_card,
    HealthcheckAnalysisCard,
    HealthcheckStatus,
)
from ax.analysis.healthcheck.regression_detection_utils import (
    detect_regressions_by_trial,
)
from ax.core.experiment import Experiment
from ax.exceptions.core import UserInputError
from ax.generation_strategy.generation_strategy import GenerationStrategy
from pyre_extensions import none_throws, override


class RegressionAnalysis(Analysis):
    r"""
    Analysis for detecting the regressing arm, metric pairs across all trials with data.
    For each metric, the regressions are defined as the arms that have a probability of
    regression above the threshold. The regression probabilities are calculated as
    posterior probabilities of the metric being above or below zero (depending if the
    metric is improving in the negative or positive direction), where the
    posteriors are with respect to the EBAshr
    (empirical Bayes Adaptive Shrinkage Model).
    """

    def __init__(self, prob_threshold: float = 0.95) -> None:
        r"""
        Args:
            prob_threshold: The threshold for the probability of metric regression.
                Regressions are defined as the arms that have a probability of
                regression above this threshold.

        """
        self.prob_threshold = prob_threshold

    @override
    def compute(
        self,
        experiment: Experiment | None,
        generation_strategy: GenerationStrategy | None = None,
        adapter: Adapter | None = None,
    ) -> HealthcheckAnalysisCard:
        r"""
        Detect the regressing arms for all trials that have data.

        Args:
            experiment: Ax experiment.
            generation_strategy: Ax generation strategy.
            adapter: Ax adapter adapter

        Returns:
            A HealthcheckAnalysisCard object with the information on regressing arms
            and the corresponding metrics the arms regress.
        """
        if experiment is None:
            raise UserInputError("Experiment cannot be None.")

        if experiment.status_quo is None:
            raise UserInputError(
                "Experiment must have a status quo arm to run the regression analysis "
                "since the regressions are relative to the status quo arm."
            )

        data = none_throws(experiment).lookup_data()
        thresholds = {
            metric: (0.0, self.prob_threshold) for metric in experiment.metrics.keys()
        }
        regressions_by_trial = detect_regressions_by_trial(
            experiment=experiment,
            thresholds=thresholds,
            data=data,
        )
        regressions_by_trial_df, regressions_msg = process_regression_dict(
            regressions_by_trial=regressions_by_trial
        )

        subtitle_base = (
            "The regression analysis health check detects arms "
            "across all trials that are regressing metrics. While metric "
            "regressions can happen (especially in exploratory rounds that use "
            "randomized parameters), users may choose to stop arms that are "
            "regressing company-critical metrics.\n\n"
        )

        if regressions_by_trial_df.shape[0] > 0:
            status = HealthcheckStatus.WARNING
            subtitle = subtitle_base + (
                "The following arms are regressing the "
                "following metrics for the respective trials: \n"
                f"{regressions_msg}"
            )

            title_status = "Warning"
        else:
            status = HealthcheckStatus.PASS
            subtitle = subtitle_base + "No metric regessions detected."
            title_status = "Success"

        return create_healthcheck_analysis_card(
            name=self.__class__.__name__,
            title=f"Ax Regression Analysis {title_status}",
            subtitle=subtitle,
            df=regressions_by_trial_df,
            status=status,
        )


def process_regression_dict(
    regressions_by_trial: dict[int, dict[str, dict[str, float]]],
) -> tuple[pd.DataFrame, str]:
    r"""
    Process the dictionary of trial indices, regressing arms and metrics into
        a dataframe and a string.

    Args:
        regressions_by_trial: A dictionary of the form
            {trial_index: {arm_name: {metric_name: probability}}}.

    Returns: A tuple containing
        - A dataFrame with columns ["trial_index", "arm_name", "metric_name",
            "probability"] and
        - A string of the form containing trial indices, regressing arms and metrics.
    """
    trial_indices = []
    arm_names = []
    metric_names = []
    probabilities = []

    msg = ""

    for trial_index, arm_metric_probs in regressions_by_trial.items():
        if arm_metric_probs is None or len(arm_metric_probs) == 0:
            continue
        msg += f"Trial {trial_index}: \n"

        for arm_name, metrics_probs in arm_metric_probs.items():
            regressing_metrics = list(metrics_probs.keys())
            metric_names.extend(regressing_metrics)
            trial_indices.extend([trial_index] * len(regressing_metrics))
            probabilities.extend(list(metrics_probs.values()))
            arm_names.extend([arm_name] * len(regressing_metrics))

            msg += f" - Arm {arm_name}: \n"
            for metric in regressing_metrics:
                msg += f"{metric}, "
            msg = msg[:-2] + " \n"

    regressions_by_trial_df = pd.DataFrame(
        {
            "trial_index": trial_indices,
            "arm_name": arm_names,
            "metric_name": metric_names,
            "probability": probabilities,
        }
    )
    if "trial_index" in regressions_by_trial_df.columns:
        regressions_by_trial_df["trial_index"] = regressions_by_trial_df[
            "trial_index"
        ].astype(int)
    return regressions_by_trial_df, msg



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from collections.abc import Mapping, Sequence
from dataclasses import dataclass, field

import torch
from ax.benchmark.benchmark_test_function import BenchmarkTestFunction


@dataclass(kw_only=True)
class IdentityTestFunction(BenchmarkTestFunction):
    """
    Test function that returns the value of parameter "x0", ignoring any others.
    """

    outcome_names: Sequence[str] = field(default_factory=lambda: ["objective"])
    n_steps: int = 1

    # pyre-fixme[14]: Inconsistent override
    def evaluate_true(self, params: Mapping[str, float]) -> torch.Tensor:
        """
        Return params["x0"] for each outcome for each time step.

        Args:
            params: A dictionary with key "x0".
        """
        value = params["x0"]
        return torch.full(
            (len(self.outcome_names), self.n_steps), value, dtype=torch.float64
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from ax.benchmark.benchmark_problem import get_soo_opt_config
from ax.core.optimization_config import OptimizationConfig
from ax.core.parameter import ParameterType, RangeParameter
from ax.core.search_space import SearchSpace


DEFAULT_METRIC_NAME: str = "Train/val_accuracy"

BASELINE_VALUES: dict[str, float] = {
    "APSFailure": 97.75948131763847,
    "Amazon_employee_access": 93.39364177908142,
    "Australian": 88.1445880383116,
    "Fashion-MNIST": 84.75904272864778,
    "KDDCup09_appetency": 96.13544312868322,
    "MiniBooNE": 85.8639428612948,
    "adult": 79.50334987749676,
    "airlines": 58.96099030718572,
    "albert": 63.885932360810884,
    "bank-marketing": 83.72755317459641,
    "blood-transfusion-service-center": 62.651717620524835,
    "car": 78.59464531457958,
    "christine": 72.22719165860138,
    "cnae-9": 92.24923138962973,
    "connect-4": 63.808749677494774,
    "covertype": 61.61393200315512,
    "credit-g": 70.45312807563056,
    "dionis": 53.71071232033245,
    "fabert": 64.44304132875557,
    "helena": 18.239085505279544,
    "higgs": 64.74999655474926,
    "jannis": 57.82155396833136,
    "jasmine": 80.48475426337272,
    "jungle_chess_2pcs_raw_endgame_complete": 65.58537332961572,
    "kc1": 77.28692486000287,
    "kr-vs-kp": 93.63368446446995,
    "mfeat-factors": 94.72758417873838,
    "nomao": 93.73968374826451,
    "numerai28.6": 51.60281273196557,
    "phoneme": 75.20979771001986,
    "segment": 78.81992685291081,
    "shuttle": 96.45744339531132,
    "sylvine": 91.15923021902736,
    "vehicle": 67.40729695042013,
    "volkert": 49.204981948803855,
}


def get_lcbench_search_space() -> SearchSpace:
    """Construct the LCBench search space."""
    search_space: SearchSpace = SearchSpace(
        parameters=[
            RangeParameter(
                name="batch_size",
                parameter_type=ParameterType.INT,
                lower=16,
                upper=512,
                log_scale=True,
            ),
            RangeParameter(
                name="max_dropout",
                parameter_type=ParameterType.FLOAT,
                lower=0.0,
                upper=1.0,  # Yes, really. Could make smaller if
                # we want to have it be more realistic.
                log_scale=False,
            ),
            RangeParameter(
                name="max_units",
                parameter_type=ParameterType.INT,
                lower=64,
                upper=1024,
                log_scale=True,
            ),
            RangeParameter(
                name="num_layers",
                parameter_type=ParameterType.INT,
                lower=1,
                upper=4,  # not a bug, even though it says 1-5 in the LCBench repo.
                # See https://github.com/automl/LCBench/issues/4
                log_scale=False,
            ),
            RangeParameter(
                name="learning_rate",
                parameter_type=ParameterType.FLOAT,
                lower=1e-4,
                upper=1e-1,
                log_scale=True,
            ),
            RangeParameter(
                name="momentum",
                parameter_type=ParameterType.FLOAT,
                lower=0.1,
                upper=0.99,
                log_scale=True,
            ),
            RangeParameter(
                name="weight_decay",
                parameter_type=ParameterType.FLOAT,
                lower=1e-5,
                upper=1e-1,
                log_scale=False,  # not a bug, see the LCBench repo.
            ),
        ]
    )
    return search_space


def get_lcbench_optimization_config(
    metric_name: str = DEFAULT_METRIC_NAME,
    observe_noise_sd: bool = False,
    use_map_metric: bool = False,
) -> OptimizationConfig:
    return get_soo_opt_config(
        outcome_names=[metric_name],
        lower_is_better=False,
        observe_noise_sd=observe_noise_sd,
        use_map_metric=use_map_metric,
    )


def get_lcbench_parameters() -> dict[str, RangeParameter]:
    return get_lcbench_search_space().parameters  # pyre-ignore [7]


def get_lcbench_parameter_names() -> list[str]:
    return list(get_lcbench_parameters().keys())


def get_lcbench_log_scale_parameter_names() -> list[str]:
    return [
        name
        for name, parameter in get_lcbench_parameters().items()
        if parameter.log_scale
    ]



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


from collections.abc import Mapping
from typing import Any

from ax.api.protocols.utils import _APIRunner
from ax.api.types import TParameterization

from ax.core.trial_status import TrialStatus
from pyre_extensions import override


class IRunner(_APIRunner):
    """
    ``Runners`` automate the process of running trials on external systems. They are
    used in conjunction with ``Metrics`` in the ``Client.run_trials`` method to
    facilitate closed-loop experimentation.
    """

    @override
    def run_trial(
        self, trial_index: int, parameterization: TParameterization
    ) -> dict[str, Any]:
        """
        Given an index and parameterization, run a trial and return a dictionary of any
        appropriate metadata. This metadata will be used to identify the trial when
        polling its status, stopping, fetching data, etc. This may hold information
        such as the trial's unique identifier on the system its running on, a
        directory where the trial is logging results to, etc.

        The metadata MUST be JSON-serializable (i.e. dict, list, str, int, float, bool,
        or None) so that Trials may be properly serialized in Ax.
        """
        ...

    @override
    def poll_trial(
        self, trial_index: int, trial_metadata: Mapping[str, Any]
    ) -> TrialStatus:
        """
        Given trial index and metadata, poll the status of the trial.
        """
        ...

    @override
    def stop_trial(
        self, trial_index: int, trial_metadata: Mapping[str, Any]
    ) -> dict[str, Any]:
        """
        Given trial index and metadata, stop the trial. Returns a dictionary of any
        appropriate metadata.

        The metadata MUST be JSON-serializable (i.e. dict, list, str, int, float, bool,
        or None) so that Trials may be properly serialized in Ax.
        """
        ...



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from unittest.mock import Mock

import pandas as pd
from ax.adapter.registry import Generators
from ax.core.arm import Arm
from ax.core.batch_trial import BatchTrial
from ax.core.data import Data
from ax.core.optimization_config import MultiObjectiveOptimizationConfig
from ax.core.trial import Trial
from ax.exceptions.core import DataRequiredError
from ax.service.utils.best_point import get_trace
from ax.service.utils.best_point_mixin import BestPointMixin
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import (
    get_experiment_with_batch_trial,
    get_experiment_with_observations,
    get_experiment_with_trial,
)
from pyre_extensions import assert_is_instance, none_throws


class TestBestPointMixin(TestCase):
    def test_get_trace(self) -> None:
        # Single objective, minimize.
        exp = get_experiment_with_observations(
            observations=[[11], [10], [9], [15], [5]], minimize=True
        )
        self.assertEqual(get_trace(exp), [11, 10, 9, 9, 5])

        # Same experiment with maximize via new optimization config.
        opt_conf = none_throws(exp.optimization_config).clone()
        opt_conf.objective.minimize = False
        self.assertEqual(get_trace(exp, opt_conf), [11, 11, 11, 15, 15])

        with self.subTest("Single objective with constraints"):
            # The second metric is the constraint and needs to be >= 0
            exp = get_experiment_with_observations(
                observations=[[11, -1], [10, 1], [9, 1], [15, -1], [11, 1]],
                minimize=False,
                constrained=True,
            )
            self.assertEqual(get_trace(exp), [float("-inf"), 10, 10, 10, 11])

            exp = get_experiment_with_observations(
                observations=[[11, -1], [10, 1], [9, 1], [15, -1], [11, 1]],
                minimize=True,
                constrained=True,
            )
            self.assertEqual(get_trace(exp), [float("inf"), 10, 9, 9, 9])

        # Scalarized.
        exp = get_experiment_with_observations(
            observations=[[1, 1], [2, 2], [3, 3]],
            scalarized=True,
        )
        self.assertEqual(get_trace(exp), [2, 4, 6])

        # Multi objective.
        exp = get_experiment_with_observations(
            observations=[[1, 1], [-1, 100], [1, 2], [3, 3], [2, 4], [2, 1]],
        )
        self.assertEqual(get_trace(exp), [1, 1, 2, 9, 11, 11])

        # W/o ObjectiveThresholds (infering ObjectiveThresholds from nadir point)
        assert_is_instance(
            exp.optimization_config, MultiObjectiveOptimizationConfig
        ).objective_thresholds = []
        self.assertEqual(get_trace(exp), [0.0, 0.0, 2.0, 8.0, 11.0, 11.0])

        # Multi-objective w/ constraints.
        exp = get_experiment_with_observations(
            observations=[[-1, 1, 1], [1, 2, 1], [3, 3, -1], [2, 4, 1], [2, 1, 1]],
            constrained=True,
        )
        self.assertEqual(get_trace(exp), [0, 2, 2, 8, 8])

        # W/ relative constraints & status quo.
        exp.status_quo = Arm(parameters={"x": 0.5, "y": 0.5}, name="status_quo")
        exp.optimization_config.outcome_constraints[0].bound = 1.0
        exp.optimization_config.outcome_constraints[0].relative = True
        # Fails if there's no data for status quo.
        with self.assertRaisesRegex(DataRequiredError, "relative constraint"):
            get_trace(exp)
        # Add data for status quo.
        trial = Trial(experiment=exp).add_arm(arm=exp.status_quo)
        df_dict = [
            {
                "trial_index": trial.index,
                "metric_name": m,
                "arm_name": "status_quo",
                "mean": 0.0,
                "sem": 0.0,
            }
            for m in ["m1", "m2", "m3"]
        ]
        status_quo_data = Data(df=pd.DataFrame.from_records(df_dict))
        exp.attach_data(data=status_quo_data)
        self.assertEqual(get_trace(exp), [0, 2, 2, 8, 8])

        # W/ first objective being minimized.
        exp = get_experiment_with_observations(
            observations=[[1, 1], [-1, 2], [3, 3], [-2, 4], [2, 1]], minimize=True
        )
        self.assertEqual(get_trace(exp), [0, 2, 2, 8, 8])

        # W/ empty data.
        exp = get_experiment_with_trial()
        self.assertEqual(get_trace(exp), [])

        # test batch trial
        exp = get_experiment_with_batch_trial()
        trial = exp.trials[0]
        exp.optimization_config.outcome_constraints[0].relative = False
        trial.mark_running(no_runner_required=True).mark_completed()
        df_dict = []
        for i, arm in enumerate(trial.arms):
            df_dict.extend(
                [
                    {
                        "trial_index": 0,
                        "metric_name": m,
                        "arm_name": arm.name,
                        "mean": float(i),
                        "sem": 0.0,
                    }
                    for m in exp.metrics.keys()
                ]
            )
        exp.attach_data(Data(df=pd.DataFrame.from_records(df_dict)))
        self.assertEqual(get_trace(exp), [len(trial.arms) - 2])
        # test that there is performance metric in the trace for each
        # completed/early-stopped trial
        trial1 = assert_is_instance(trial, BatchTrial).clone_to()
        trial1.mark_abandoned(unsafe=True)
        trial2 = exp.new_batch_trial(Generators.SOBOL(experiment=exp).gen(n=3))
        trial2.mark_running(no_runner_required=True).mark_completed()
        df_dict2 = []
        for i, arm in enumerate(trial2.arms):
            df_dict2.extend(
                [
                    {
                        "trial_index": 2,
                        "metric_name": m,
                        "arm_name": arm.name,
                        "mean": 10 * float(i),
                        "sem": 0.0,
                    }
                    for m in exp.metrics.keys()
                ]
            )
        exp.attach_data(Data(df=pd.DataFrame.from_records(df_dict2)))
        self.assertEqual(get_trace(exp), [2, 20.0])

    def test_get_hypervolume(self) -> None:
        # W/ empty data.
        exp = get_experiment_with_trial()
        self.assertEqual(BestPointMixin._get_hypervolume(exp, Mock()), 0.0)

    def test_get_best_observed_value(self) -> None:
        # Alias for easier access.
        get_best = BestPointMixin._get_best_observed_value

        # Single objective, minimize.
        exp = get_experiment_with_observations(
            observations=[[11], [10], [9], [15], [5]], minimize=True
        )
        self.assertEqual(get_best(exp), 5)
        # Same experiment with maximize via new optimization config.
        opt_conf = none_throws(exp.optimization_config).clone()
        opt_conf.objective.minimize = False
        self.assertEqual(get_best(exp, opt_conf), 15)

        # Scalarized.
        exp = get_experiment_with_observations(
            observations=[[1, 1], [2, 2], [3, 3]],
            scalarized=True,
        )
        self.assertEqual(get_best(exp), 6)

        # Exclude out of design arms
        exp = get_experiment_with_observations(
            observations=[[11], [10], [9], [15], [5]],
            parameterizations=[
                {"x": 0.0, "y": 0.0},
                {"x": 0.1, "y": 0.0},
                {"x": 10.0, "y": 10.0},  # out of design
                {"x": 0.2, "y": 0.0},
                {"x": 10.1, "y": 10.0},  # out of design
            ],
            minimize=True,
        )
        self.assertEqual(get_best(exp), 10)  # 5 and 9 are out of design



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import math
from typing import Optional, TYPE_CHECKING

import numpy as np
from ax.adapter.data_utils import ExperimentData
from ax.adapter.transforms.base import Transform
from ax.core.observation import Observation, ObservationFeatures
from ax.core.parameter import ParameterType, RangeParameter
from ax.core.search_space import SearchSpace
from ax.generators.types import TConfig
from pyre_extensions import assert_is_instance

if TYPE_CHECKING:
    # import as module to make sphinx-autodoc-typehints happy
    from ax import adapter as adapter_module  # noqa F401


class Log(Transform):
    """Apply log base 10 to a float RangeParameter domain.

    Transform is done in-place.
    """

    def __init__(
        self,
        search_space: SearchSpace | None = None,
        observations: list[Observation] | None = None,
        experiment_data: ExperimentData | None = None,
        adapter: Optional["adapter_module.base.Adapter"] = None,
        config: TConfig | None = None,
    ) -> None:
        assert search_space is not None, "Log requires search space"
        super().__init__(
            search_space=search_space,
            observations=observations,
            experiment_data=experiment_data,
            adapter=adapter,
            config=config,
        )
        # Identify parameters that should be transformed
        self.transform_parameters: set[str] = {
            p_name
            for p_name, p in search_space.parameters.items()
            if isinstance(p, RangeParameter)
            and p.parameter_type == ParameterType.FLOAT
            and p.log_scale is True
        }

    def transform_observation_features(
        self, observation_features: list[ObservationFeatures]
    ) -> list[ObservationFeatures]:
        for obsf in observation_features:
            for p_name in self.transform_parameters:
                if p_name in obsf.parameters:
                    param: float = assert_is_instance(obsf.parameters[p_name], float)
                    obsf.parameters[p_name] = math.log10(param)
        return observation_features

    def _transform_search_space(self, search_space: SearchSpace) -> SearchSpace:
        for p_name, p in search_space.parameters.items():
            if p_name in self.transform_parameters and isinstance(p, RangeParameter):
                # Don't round in log space
                if p.digits is not None:
                    p.set_digits(digits=None)
                p.set_log_scale(False).update_range(
                    lower=math.log10(p.lower), upper=math.log10(p.upper)
                )
                if p.target_value is not None:
                    p._target_value = math.log10(p.target_value)  # pyre-ignore [6]
        return search_space

    def untransform_observation_features(
        self, observation_features: list[ObservationFeatures]
    ) -> list[ObservationFeatures]:
        for obsf in observation_features:
            for p_name in self.transform_parameters:
                if p_name in obsf.parameters:
                    param: float = assert_is_instance(obsf.parameters[p_name], float)
                    obsf.parameters[p_name] = math.pow(10, param)
        return observation_features

    def transform_experiment_data(
        self, experiment_data: ExperimentData
    ) -> ExperimentData:
        arm_data = experiment_data.arm_data
        for p_name in self.transform_parameters:
            arm_data[p_name] = np.log10(arm_data[p_name])
        return ExperimentData(
            arm_data=arm_data, observation_data=experiment_data.observation_data
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


from logging import Logger
from unittest.mock import patch

from ax.utils.common.kwargs import warn_on_kwargs
from ax.utils.common.logger import get_logger
from ax.utils.common.testutils import TestCase

logger: Logger = get_logger("ax.utils.common.kwargs")


class TestWarnOnKwargs(TestCase):
    def test_it_warns_if_kwargs_are_passed(self) -> None:
        with patch.object(logger, "warning") as mock_warning:

            def callable_arg() -> None:
                return

            warn_on_kwargs(callable_with_kwargs=callable_arg, foo="")
        mock_warning.assert_called_once_with(
            "Found unexpected kwargs: %s while calling %s "
            "from JSON. These kwargs will be ignored.",
            {"foo": ""},
            callable_arg,
        )

    def test_it_does_not_warn_if_no_kwargs_are_passed(self) -> None:
        with patch.object(logger, "warning") as mock_warning:
            warn_on_kwargs(callable_with_kwargs=lambda: None)
        mock_warning.assert_not_called()



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

import inspect
from collections import OrderedDict
from collections.abc import Sequence
from copy import deepcopy
from dataclasses import dataclass, field
from logging import Logger
from typing import Any, Mapping

import numpy as np
import torch
from ax.core.search_space import SearchSpaceDigest
from ax.core.types import TCandidateMetadata
from ax.exceptions.core import AxError, UnsupportedError, UserInputError
from ax.exceptions.model import ModelError
from ax.generators.model_utils import best_in_sample_point
from ax.generators.torch.botorch_modular.input_constructors.covar_modules import (
    covar_module_argparse,
)
from ax.generators.torch.botorch_modular.input_constructors.input_transforms import (
    input_transform_argparse,
)
from ax.generators.torch.botorch_modular.input_constructors.outcome_transform import (
    outcome_transform_argparse,
)
from ax.generators.torch.botorch_modular.utils import (
    convert_to_block_design,
    copy_model_config_with_default_values,
    fit_botorch_model,
    get_cv_fold,
    ModelConfig,
    subset_state_dict,
    use_model_list,
)
from ax.generators.torch.utils import (
    _to_inequality_constraints,
    pick_best_out_of_sample_point_acqf_class,
    predict_from_model,
)
from ax.generators.torch_base import TorchOptConfig
from ax.generators.types import TConfig
from ax.utils.common.base import Base
from ax.utils.common.constants import Keys
from ax.utils.common.logger import get_logger
from ax.utils.common.typeutils import (
    _argparse_type_encoder,
    assert_is_instance_optional,
)
from ax.utils.stats.model_fit_stats import (
    DIAGNOSTIC_FN_DIRECTIONS,
    DIAGNOSTIC_FNS,
    ModelFitMetricDirection,
    RANK_CORRELATION,
)
from botorch.exceptions.errors import ModelFittingError
from botorch.models.gpytorch import GPyTorchModel
from botorch.models.model import Model, ModelList
from botorch.models.model_list_gp_regression import ModelListGP
from botorch.models.multitask import MultiTaskGP
from botorch.models.transforms.input import (
    ChainedInputTransform,
    InputPerturbation,
    InputTransform,
    Normalize,
)
from botorch.models.transforms.outcome import ChainedOutcomeTransform, OutcomeTransform
from botorch.posteriors.fully_bayesian import GaussianMixturePosterior
from botorch.posteriors.gpytorch import GPyTorchPosterior
from botorch.settings import validate_input_scaling
from botorch.utils.containers import SliceContainer
from botorch.utils.datasets import MultiTaskDataset, RankingDataset, SupervisedDataset
from botorch.utils.dispatcher import Dispatcher
from botorch.utils.evaluation import AIC, BIC, compute_in_sample_model_fit_metric, MLL
from botorch.utils.transforms import normalize_indices
from botorch.utils.types import _DefaultType
from gpytorch.models.exact_gp import ExactGP
from pyre_extensions import assert_is_instance, none_throws
from torch import Tensor
from torch.nn import Module

NOT_YET_FIT_MSG = (
    "Underlying BoTorch `Model` has not yet received its training_data. "
    "Please fit the model first."
)

logger: Logger = get_logger(__name__)
MODEL_SELECTION_METRIC_DIRECTIONS: dict[str, ModelFitMetricDirection] = {
    **DIAGNOSTIC_FN_DIRECTIONS,
    MLL: ModelFitMetricDirection.MAXIMIZE,
    AIC: ModelFitMetricDirection.MINIMIZE,
    BIC: ModelFitMetricDirection.MINIMIZE,
}


def _extract_model_kwargs(
    search_space_digest: SearchSpaceDigest, botorch_model_class: type[Model]
) -> dict[str, list[int] | dict[int, dict[int, list[int]]] | int]:
    """
    Extracts keyword arguments that are passed to the `construct_inputs`
    method of a BoTorch `Model` class.

    Args:
        search_space_digest: A `SearchSpaceDigest`.
        botorch_model_class: The BoTorch model class to extract kwargs for.

    Returns:
        A dict of fidelity features, categorical features, and, if present, task
        features.
    """
    fidelity_features = search_space_digest.fidelity_features
    task_features = search_space_digest.task_features
    if len(fidelity_features) > 0 and len(task_features) > 0:
        raise NotImplementedError(
            "Multi-Fidelity GP models with task_features are "
            "currently not supported."
        )
    if len(task_features) > 1:
        raise NotImplementedError("Multiple task features are not supported.")
    elif len(task_features) == 0 and issubclass(botorch_model_class, MultiTaskGP):
        # This is handled in Surrogate.model_selection and the MTGP will be
        # skipped if there is no task feature.
        raise ModelFittingError("Cannot fit MultiTaskGP without task feature.")

    kwargs: dict[str, list[int] | dict[int, dict[int, list[int]]] | int] = {}
    if len(search_space_digest.categorical_features) > 0:
        kwargs["categorical_features"] = search_space_digest.categorical_features
    if len(fidelity_features) > 0:
        kwargs["fidelity_features"] = fidelity_features
    if len(task_features) == 1:
        task_feature = task_features[0]
        if task_feature == len(search_space_digest.bounds) - 1:
            # to support heterogeneous search spaces
            task_feature = -1
        kwargs["task_feature"] = task_feature
    # Regular BoTorch models do not expect the argument `hierarchical_dependencies`.
    # For now, it is the user's responsibility to make sure a hierarchical model is used
    # when the HSS is not flattened.
    if search_space_digest.hierarchical_dependencies:
        kwargs["hierarchical_dependencies"] = (
            search_space_digest.hierarchical_dependencies
        )
    return kwargs


def _make_botorch_input_transform(
    input_transform_classes: list[type[InputTransform]] | _DefaultType,
    input_transform_options: dict[str, dict[str, Any]],
    dataset: SupervisedDataset,
    search_space_digest: SearchSpaceDigest,
) -> InputTransform | None:
    """
    Makes a BoTorch input transform from the provided input classes and options.
    """
    if isinstance(input_transform_classes, _DefaultType):
        transforms = _construct_default_input_transforms(
            search_space_digest=search_space_digest, dataset=dataset
        )
    else:
        transforms = _construct_specified_input_transforms(
            input_transform_classes=input_transform_classes,
            dataset=dataset,
            search_space_digest=search_space_digest,
            input_transform_options=input_transform_options,
        )
    if len(transforms) == 0:
        return None
    elif len(transforms) > 1:
        return ChainedInputTransform(
            **{f"tf{i}": t_i for i, t_i in enumerate(transforms)}
        )
    else:
        return transforms[0]


def _construct_specified_input_transforms(
    input_transform_classes: list[type[InputTransform]],
    input_transform_options: dict[str, dict[str, Any]],
    dataset: SupervisedDataset,
    search_space_digest: SearchSpaceDigest,
) -> list[InputTransform]:
    """Constructs a list of input transforms from input transform classes and
    options provided in ``ModelConfig``.
    """
    if not (
        isinstance(input_transform_classes, list)
        and all(issubclass(c, InputTransform) for c in input_transform_classes)
    ):
        raise UserInputError(
            "Expected a list of input transform classes. "
            f"Got {input_transform_classes=}."
        )
    if search_space_digest.robust_digest is not None:
        input_transform_classes = [InputPerturbation] + input_transform_classes

    input_transform_kwargs = [
        input_transform_argparse(
            transform_class,
            dataset=dataset,
            search_space_digest=search_space_digest,
            input_transform_options=deepcopy(  # In case of in-place modifications.
                input_transform_options.get(transform_class.__name__, {})
            ),
        )
        for transform_class in input_transform_classes
    ]

    return [
        # pyre-fixme[45]: Cannot instantiate abstract class `InputTransform`.
        transform_class(**single_input_transform_kwargs)
        for transform_class, single_input_transform_kwargs in zip(
            input_transform_classes, input_transform_kwargs
        )
    ]


def _construct_default_input_transforms(
    search_space_digest: SearchSpaceDigest,
    dataset: SupervisedDataset,
) -> list[InputTransform]:
    """Construct the default input transforms for the given search space digest.

    The default transforms are added in this order:
    - If the search space digest has a robust digest, an ``InputPerturbation`` transform
        is used.
    - If the bounds for the non-task features are not [0, 1], a ``Normalize`` transform
        is used. The transfrom only applies to the non-task features.
    """
    transforms = []
    # Add InputPerturbation if there is a robust digest.
    if search_space_digest.robust_digest is not None:
        transforms.append(
            InputPerturbation(
                **input_transform_argparse(
                    InputPerturbation,
                    dataset=dataset,
                    search_space_digest=search_space_digest,
                )
            )
        )
    # Processing for Normalize.
    input_transform_options = input_transform_argparse(
        Normalize,
        dataset=dataset,
        search_space_digest=search_space_digest,
    )
    bounds = input_transform_options.get("bounds")
    indices = input_transform_options.get("indices")
    # Skip the Normalize transform if the bounds are [0, 1].
    if bounds is not None:
        if indices is not None:
            bounds = bounds[:, indices]
        lower_bounds, upper_bounds = bounds
        if torch.allclose(
            lower_bounds, torch.zeros_like(lower_bounds)
        ) and torch.allclose(upper_bounds, torch.ones_like(upper_bounds)):
            return transforms
    transforms.append(Normalize(**input_transform_options))
    return transforms


def _make_botorch_outcome_transform(
    outcome_transform_classes: list[type[OutcomeTransform]],
    outcome_transform_options: dict[str, dict[str, Any]],
    dataset: SupervisedDataset,
) -> OutcomeTransform | None:
    """
    Makes a BoTorch outcome transform from the provided classes and options.
    """
    if not (
        isinstance(outcome_transform_classes, list)
        and all(issubclass(c, OutcomeTransform) for c in outcome_transform_classes)
    ):
        raise UserInputError("Expected a list of outcome transforms.")
    if len(outcome_transform_classes) == 0:
        return None

    outcome_transform_kwargs = [
        outcome_transform_argparse(
            transform_class,
            outcome_transform_options=deepcopy(  # In case of in-place modifications.
                outcome_transform_options.get(transform_class.__name__, {})
            ),
            dataset=dataset,
        )
        for transform_class in outcome_transform_classes
    ]

    outcome_transforms = [
        # pyre-fixme[45]: Cannot instantiate abstract class `OutcomeTransform`.
        transform_class(**single_outcome_transform_kwargs)
        for transform_class, single_outcome_transform_kwargs in zip(
            outcome_transform_classes, outcome_transform_kwargs
        )
    ]

    outcome_transform_instance = (
        ChainedOutcomeTransform(
            **{f"otf{i}": otf for i, otf in enumerate(outcome_transforms)}
        )
        if len(outcome_transforms) > 1
        else outcome_transforms[0]
    )
    return outcome_transform_instance


def _construct_submodules(
    model_config: ModelConfig,
    dataset: SupervisedDataset,
    search_space_digest: SearchSpaceDigest,
    botorch_model_class: type[Model],
) -> dict[str, Module | None]:
    """Constructs the submodules for the BoTorch model from the inputs
    extracted from the ``ModelConfig``. If the corresponding inputs are
    specified, the `covar_module`, `likelihood`, `input_transform`, and
    `outcome_transform` submodules are constructed.
    """
    botorch_model_class_args: list[str] = inspect.getfullargspec(
        botorch_model_class
    ).args

    def _error_if_arg_not_supported(arg_name: str) -> None:
        if arg_name not in botorch_model_class_args:
            raise UserInputError(
                f"The BoTorch model class {botorch_model_class.__name__} does not "
                f"support the input {arg_name}."
            )

    submodules: dict[str, Module | None] = {}
    # NOTE: Using the walrus operator here and below helps pyre.
    if (covar_class := model_config.covar_module_class) is not None:
        _error_if_arg_not_supported("covar_module")
        covar_module_kwargs = covar_module_argparse(
            covar_class,
            dataset=dataset,
            botorch_model_class=botorch_model_class,
            **deepcopy(model_config.covar_module_options),
        )
        # pyre-ignore [45]: Cannot instantiate abstract class `Kernel`.
        submodules["covar_module"] = covar_class(**covar_module_kwargs)

    if (likelihood_class := model_config.likelihood_class) is not None:
        _error_if_arg_not_supported("likelihood")
        # pyre-ignore [45]: Cannot instantiate abstract class `Likelihood`.
        submodules["likelihood"] = likelihood_class(
            **deepcopy(model_config.likelihood_options)
        )

    if (
        input_transform_classes := model_config.input_transform_classes
    ) is not None or search_space_digest.robust_digest is not None:
        _error_if_arg_not_supported("input_transform")
        submodules["input_transform"] = _make_botorch_input_transform(
            input_transform_classes=input_transform_classes or [],
            input_transform_options=model_config.input_transform_options or {},
            dataset=dataset,
            search_space_digest=search_space_digest,
        )

    if (
        outcome_transform_classes := model_config.outcome_transform_classes
    ) is not None:
        _error_if_arg_not_supported("outcome_transform")
        submodules["outcome_transform"] = _make_botorch_outcome_transform(
            outcome_transform_classes=outcome_transform_classes,
            outcome_transform_options=model_config.outcome_transform_options or {},
            dataset=dataset,
        )
    elif "outcome_transform" in botorch_model_class_args:
        # This is a temporary solution until all BoTorch models use
        # `Standardize` by default, see TODO [T197435440].
        # After this, we should update `Surrogate` to use `DEFAULT`
        # (https://fburl.com/code/22f4397e) for both of these args. This will
        # allow users to explicitly disable the default transforms by passing
        # in `None`.
        submodules["outcome_transform"] = None

    return submodules


@dataclass(frozen=True)
class SurrogateSpec:
    """
    Fields in the SurrogateSpec dataclass correspond to arguments in
    ``Surrogate.__init__``, except for ``outcomes`` which is used to specify which
    outcomes the Surrogate is responsible for modeling.
    When ``BoTorchGenerator.fit`` is called, these fields will be used to construct the
    requisite Surrogate objects.
    If ``outcomes`` is left empty then no outcomes will be fit to the Surrogate.

    Args:
        model_configs: List of model configs. Each model config is a specification of
            a surrogate model. Defaults to a single ``ModelConfig`` with all defaults.
        metric_to_model_configs: Dictionary mapping metric names to a list of model
            configs for that metric.
        eval_criterion: The name of the evaluation criteria to use. These are defined in
            ``ax.utils.stats.model_fit_stats``. Defaults to rank correlation.
        outcomes: List of outcomes names.
        use_posterior_predictive: Whether to use posterior predictive in
            cross-validation.
        num_folds: The number of folds to use in cross-validation. If None, then
            leave-one-out.
    """

    model_configs: list[ModelConfig] = field(default_factory=lambda: [ModelConfig()])
    metric_to_model_configs: dict[str, list[ModelConfig]] = field(default_factory=dict)
    eval_criterion: str = RANK_CORRELATION
    outcomes: list[str] = field(default_factory=list)
    allow_batched_models: bool = True
    use_posterior_predictive: bool = False
    num_folds: int | None = 10


class Surrogate(Base):
    """
    **All classes in 'botorch_modular' directory are under
    construction, incomplete, and should be treated as alpha
    versions only.**

    Ax wrapper for BoTorch ``Model``, subcomponent of ``BoTorchGenerator``
    and is not meant to be used outside of it.

    Args:
        surrogate_spec: A ``SurrogateSpec`` that specifies the option to use when
            constructing the surrogate models. See the docstring of ``SurrogateSpec``
            for supported options and ``ModelConfig`` for additional details.
        allow_batched_models: Set to true to fit the models in a batch if supported.
            Set to false to fit individual models to each metric in a loop.
        refit_on_cv: Whether to refit the model on the cross-validation folds.
        warm_start_refit: Whether to warm-start refitting from the current state_dict
            during cross-validation. If refit_on_cv is True, generally one
            would set this to be False, so that no information is leaked between or
            across folds.
        metric_to_best_model_config: Dictionary mapping a metric name to the best
            model config. This is only used by `BoTorchGenerator.cross_validate` and
            for logging what model was used.

    """

    def __init__(
        self,
        surrogate_spec: SurrogateSpec | None = None,
        allow_batched_models: bool = True,
        refit_on_cv: bool = False,
        warm_start_refit: bool = True,
        metric_to_best_model_config: dict[str, ModelConfig] | None = None,
    ) -> None:
        if surrogate_spec is None:
            surrogate_spec = SurrogateSpec(allow_batched_models=allow_batched_models)

        self.surrogate_spec: SurrogateSpec = surrogate_spec
        # Store the last dataset used to fit the model for a given metric(s).
        # If the new dataset is identical, we will skip model fitting for that metric.
        # The keys are `tuple(dataset.outcome_names)`.
        self._last_datasets: dict[tuple[str], SupervisedDataset] = {}
        # Store a reference from a tuple of metric names to the BoTorch Model
        # corresponding to those metrics. In most cases this will be a one-tuple,
        # though we need n-tuples for LCE-M models. This will be used to skip model
        # construction & fitting if the datasets are identical.
        self._submodels: dict[tuple[str], Model] = {}
        self.metric_to_best_model_config: dict[str, ModelConfig] = (
            metric_to_best_model_config or {}
        )
        # Store a reference to search space digest used while fitting the cached models.
        # We will re-fit the models if the search space digest changes.
        self._last_search_space_digest: SearchSpaceDigest | None = None

        # These are later updated during model fitting.
        self._training_data: list[SupervisedDataset] | None = None
        self._outcomes: list[str] | None = None
        self._model: Model | None = None
        self.refit_on_cv = refit_on_cv
        self.warm_start_refit = warm_start_refit
        # Updated during model selection
        self._model_name_to_eval: dict[str, dict[str, float]] = {}
        self._model_name_to_model: dict[str, dict[str, Model]] = {}

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}" f" surrogate_spec={self.surrogate_spec}>"

    @property
    def model(self) -> Model:
        if self._model is None:
            raise ModelError(
                "BoTorch `Model` has not yet been constructed, please fit the "
                "surrogate first (done via `BoTorchGenerator.fit`)."
            )
        return self._model

    @property
    def training_data(self) -> list[SupervisedDataset]:
        if self._training_data is None:
            raise ModelError(NOT_YET_FIT_MSG)
        return self._training_data

    @property
    def Xs(self) -> list[Tensor]:
        # Handles multi-output models. TODO: Improve this!
        training_data = self.training_data
        Xs = []
        for dataset in training_data:
            if isinstance(dataset, RankingDataset):
                # directly accessing the d-dim X tensor values
                # instead of the augmented 2*d-dim dataset.X from RankingDataset
                Xi = assert_is_instance(
                    dataset._X,
                    SliceContainer,
                ).values
            else:
                Xi = dataset.X
            for _ in range(dataset.Y.shape[-1]):
                Xs.append(Xi)
        return Xs

    @property
    def dtype(self) -> torch.dtype:
        return self.training_data[0].X.dtype

    @property
    def device(self) -> torch.device:
        return self.training_data[0].X.device

    def clone_reset(self) -> Surrogate:
        return self.__class__(**self._serialize_attributes_as_kwargs())

    def _construct_model(
        self,
        dataset: SupervisedDataset,
        search_space_digest: SearchSpaceDigest,
        model_config: ModelConfig,
        state_dict: Mapping[str, Tensor] | None,
        refit: bool,
    ) -> Model:
        """Constructs the underlying BoTorch ``Model`` using the training data.

        If the dataset and model class are identical to those used while training
        the cached sub-model, we skip model fitting and return the cached model.

        Args:
            dataset: Training data for the model (for one outcome for
                the default `Surrogate`, with the exception of batched
                multi-output case, where training data is formatted with just
                one X and concatenated Ys).
            search_space_digest: Search space digest used to set up model arguments.
            model_config: The model_config.
            default_botorch_model_class: The default ``Model`` class to be used as the
                underlying BoTorch model, if the model_config does not specify one.
            state_dict: Optional state dict to load. This should be subsetted for
                the current submodel being constructed.
            refit: Whether to re-optimize model parameters.
        """
        outcome_names = tuple(dataset.outcome_names)
        # Fill in default values for model_configs given dataset
        model_config = copy_model_config_with_default_values(
            model_config=model_config,
            dataset=dataset,
            search_space_digest=search_space_digest,
        )
        botorch_model_class = none_throws(model_config.botorch_model_class)
        if self._dataset_matches_cache(dataset=dataset):
            return self._submodels[outcome_names]
        formatted_model_inputs = submodel_input_constructor(
            botorch_model_class,  # Do not pass as kwarg since this is used to dispatch.
            model_config=model_config,
            dataset=dataset,
            search_space_digest=search_space_digest,
            surrogate=self,
        )
        # pyre-ignore [45]
        model = botorch_model_class(**formatted_model_inputs)
        if state_dict is not None and (not refit or self.warm_start_refit):
            model.load_state_dict(state_dict)
        if state_dict is None or refit:
            fit_botorch_model(
                model,  # Intentionally not using named args for dispatcher.
                mll_class=model_config.mll_class,
                mll_options=model_config.mll_options,
            )

        return model

    def _dataset_matches_cache(
        self,
        dataset: SupervisedDataset,
    ) -> bool:
        """Returns `True` if the given dataset matches the last dataset used to fit
        the model for the corresponding outcomes.
        """
        outcome_names = tuple(dataset.outcome_names)
        return (
            outcome_names in self._submodels
            and dataset == self._last_datasets[outcome_names]
        )

    def fit(
        self,
        datasets: Sequence[SupervisedDataset],
        search_space_digest: SearchSpaceDigest,
        candidate_metadata: list[list[TCandidateMetadata]] | None = None,
        state_dict: Mapping[str, Tensor] | None = None,
        refit: bool = True,
        repeat_model_selection_if_dataset_changed: bool = True,
    ) -> None:
        """Fits the underlying BoTorch ``Model`` to ``m`` outcomes.

        NOTE: ``state_dict`` and ``refit`` keyword arguments control how the
        underlying BoTorch ``Model`` will be fit: whether its parameters will
        be reoptimized and whether it will be warm-started from a given state.

        There are three possibilities:

        * ``fit(state_dict=None)``: fit model from scratch (optimize model
          parameters and set its training data used for inference),
        * ``fit(state_dict=some_state_dict, refit=True)``: warm-start refit
          with a state dict of parameters (still re-optimize model parameters
          and set the training data),
        * ``fit(state_dict=some_state_dict, refit=False)``: load model parameters
          without refitting, but set new training data (used in cross-validation,
          for example).

        Args:
            datasets: A list of ``SupervisedDataset`` containers, each
                corresponding to the data of one metric (outcome), to be passed
                to ``Model.construct_inputs`` in BoTorch.
            search_space_digest: A ``SearchSpaceDigest`` object containing
                metadata on the features in the datasets.
            candidate_metadata: Model-produced metadata for candidates, in
                the order corresponding to the Xs.
            state_dict: Optional state dict to load.
            refit: Whether to re-optimize model parameters.
            repeat_model_selection_if_dataset_changed: Whether to repeat model
                selection, ignoring previously found best config, if the dataset
                for the corresponding outcomes has changed. This is typically
                set to `True` when called from ``BoTorchGenerator.fit`` but set
                to `False` when called from ``BoTorchGenerator.cross_validate``.
                During cross_validation, we want to evaluate the quality of the
                previously selected best model, rather than repeating model selection
                for each fold.
        """
        self._discard_cached_model_and_data_if_search_space_digest_changed(
            search_space_digest=search_space_digest
        )
        # Deepcopy so that we are not making in-place changes when fill default values
        metric_to_model_configs = deepcopy(self.surrogate_spec.metric_to_model_configs)

        # To determine whether to use ModelList under the hood, we need to check for
        # the batched multi-output case, so we first see which model would be chosen
        # given the Yvars and the properties of data.
        should_use_model_list = use_model_list(
            datasets=datasets,
            search_space_digest=search_space_digest,
            model_configs=self.surrogate_spec.model_configs,
            allow_batched_models=self.surrogate_spec.allow_batched_models,
            metric_to_model_configs=metric_to_model_configs,
        )

        if not should_use_model_list and len(datasets) > 1:
            try:
                datasets = convert_to_block_design(datasets=datasets, force=False)
            except UnsupportedError as e:
                # If the block design conversion fails, use model-list.
                logger.warning(
                    "Conversion to block design failed. Using model-list instead."
                    f"Original error: {e}"
                )
                should_use_model_list = True
        self._training_data = list(datasets)  # So that it can be modified if needed.

        feature_names_set = set(search_space_digest.feature_names)
        models = []
        outcome_names = []
        for i, dataset in enumerate(datasets):
            submodel_state_dict = None
            if state_dict is not None:
                if should_use_model_list:
                    submodel_state_dict = subset_state_dict(
                        state_dict=state_dict, submodel_index=i
                    )
                else:
                    submodel_state_dict = state_dict
            outcome_name_tuple = tuple(dataset.outcome_names)
            first_outcome_name = outcome_name_tuple[0]

            # If no model config is specified for the (likely aux) preference dataset,
            # fallback to the default model config to avoid model fitting failure
            if (
                isinstance(dataset, RankingDataset)
                and first_outcome_name not in metric_to_model_configs
            ):
                metric_to_model_configs[first_outcome_name] = [ModelConfig()]

            model_configs = (
                metric_to_model_configs[first_outcome_name]
                if first_outcome_name in metric_to_model_configs
                else self.surrogate_spec.model_configs
            )
            # Case 1: There is either 1 model config, or we don't want to re-do
            # model selection and we know what the previous best model was.
            if (
                not repeat_model_selection_if_dataset_changed
                or self._dataset_matches_cache(dataset=dataset)
            ):
                # Re-use the best model config, if the dataset hasn't changed or
                # `repeat_model_selection_if_dataset_changed` is set to `False`.
                model_config = self.metric_to_best_model_config.get(first_outcome_name)
            else:
                model_config = None
            # Model selection is not performed if the best `ModelConfig` has already
            # been identified (as specified in `metric_to_best_model_config`).
            # The reason for doing this is to support the following flow:
            # - Fit model to data and perform model selection, refitting on each fold
            #   if `refit_on_cv=True`. This will set the best ModelConfig in
            #   metric_to_best_model_config.
            # - Evaluate the choice of model/visualize its performance via
            #   `Adapter.cross_validate``. This also will refit on each fold if
            #   `refit_on_cv=True`, but we wouldn't want to perform model selection
            #   on each fold, but rather show the performance of the selecting
            #   `ModelConfig`` since that is what will be used.
            if len(model_configs) == 1 or (model_config is not None):
                best_model_config = model_config or model_configs[0]
                model = self._construct_model(
                    dataset=dataset,
                    search_space_digest=search_space_digest,
                    model_config=best_model_config,
                    state_dict=submodel_state_dict,
                    refit=refit,
                )
            # Case 2: There is more than 1 model config and we want to refit
            # or don't know what the previous best model was
            else:
                if len(dataset.outcome_names) > 1:
                    raise UnsupportedError(
                        "Multiple model configs are not supported with datasets that"
                        " contain multiple outcomes. Each dataset must contain only "
                        "one outcome."
                    )
                model, best_model_config = self.model_selection(
                    dataset=dataset,
                    model_configs=model_configs,
                    search_space_digest=search_space_digest,
                    candidate_metadata=candidate_metadata,
                )

            # Only update the outcome names and models if the dataset input matches
            # the feature names from the search space digest. Otherwise we only
            # keep the model within self._submodels as it may be models fitted on
            # auxiliary data such as the preference model for BOPE
            if set(dataset.feature_names) == feature_names_set:
                models.append(model)
                outcome_names.extend(dataset.outcome_names)

            # store best model config, model, and dataset
            for metric_name in dataset.outcome_names:
                self.metric_to_best_model_config[metric_name] = none_throws(
                    best_model_config
                )
            self._submodels[outcome_name_tuple] = model
            self._last_datasets[outcome_name_tuple] = dataset

        if should_use_model_list:
            if all(isinstance(model, GPyTorchModel) for model in models):
                self._model = ModelListGP(*models)
            else:
                self._model = ModelList(*models)
        else:
            self._model = models[0]
        self._outcomes = outcome_names  # In the order of input datasets

    def model_selection(
        self,
        dataset: SupervisedDataset,
        model_configs: list[ModelConfig],
        search_space_digest: SearchSpaceDigest,
        candidate_metadata: list[list[TCandidateMetadata]] | None = None,
    ) -> tuple[Model, ModelConfig]:
        """Perform model selection over a list of model configs.

        This selects the best botorch Model across the provided model configs
        based on the SurrogateSpec's eval_criteria. The eval_criteria is
        computed using LOOCV on the provided dataset. The best model config is saved
        in self.metric_to_best_model_config for future use (e.g. for using cross-
        validation at the Adapter level).

        Args:
            dataset: Training data for the model
            model_configs: The model_configs.
            default_botorch_model_class: The default ``Model`` class to be used as
                the default, if no botorch_model_class is specified in the
                    model_config.
            search_space_digest: Search space digest.
            candidate_metadata: Model-produced metadata for candidates.

        Returns:
            A two element tuple containing:
                - The best model according to the eval_criterion.
                - The ModelConfig for the best model.

        """
        if (
            isinstance(dataset, MultiTaskDataset)
            and assert_is_instance(dataset, MultiTaskDataset).has_heterogeneous_features
        ):
            raise UnsupportedError(
                "Model selection is not supported for datasets with heterogeneous "
                "features."
            )
        # loop over model configs, fit model for each config, perform LOOCV, select
        # best model according to specified criterion
        maximize = (
            MODEL_SELECTION_METRIC_DIRECTIONS[self.surrogate_spec.eval_criterion]
            == ModelFitMetricDirection.MAXIMIZE
        )
        prefix = "-" if maximize else ""
        best_eval_metric = float(f"{prefix}inf")
        best_model = None
        best_model_config = None
        outcome_name = dataset.outcome_names[0]
        self._model_name_to_eval[outcome_name] = {}
        self._model_name_to_model[outcome_name] = {}
        for model_config in model_configs:
            # fit model to all data
            try:
                model = self._construct_model(
                    dataset=dataset,
                    search_space_digest=search_space_digest,
                    model_config=model_config,
                    state_dict=None,
                    refit=True,
                )
                state_dict = model.state_dict()
                # perform LOOCV
                if self.surrogate_spec.eval_criterion in (AIC, BIC, MLL):
                    eval_metric = compute_in_sample_model_fit_metric(
                        model=assert_is_instance(model, ExactGP),
                        criterion=self.surrogate_spec.eval_criterion,
                    )
                else:
                    eval_metric = self.cross_validate(
                        dataset=dataset,
                        search_space_digest=search_space_digest,
                        model_config=model_config,
                        # pyre-fixme [6]: In call `Surrogate.cross_validate`, for
                        # argument  `state_dict`, expected `Optional[OrderedDict[str,
                        # Tensor]]` but got `Dict[str, typing.Any]`.
                        state_dict=state_dict,
                    )
            except ModelFittingError as e:
                logger.warning(
                    f"Model {model_config} failed to fit with error {e}. Skipping."
                )
                continue
            self._model_name_to_eval[outcome_name][model_config.identifier] = (
                eval_metric
            )
            self._model_name_to_model[outcome_name][model_config.identifier] = model
            if maximize ^ (eval_metric < best_eval_metric):
                best_eval_metric = eval_metric
                best_model = model
                best_model_config = model_config
        if best_model is None:
            raise AxError(
                "No model configs were able to fit the data. Please check your "
                "model configs and/or data."
            )
        return none_throws(best_model), none_throws(best_model_config)

    def cross_validate(
        self,
        dataset: SupervisedDataset,
        model_config: ModelConfig,
        search_space_digest: SearchSpaceDigest,
        state_dict: OrderedDict[str, Tensor] | None = None,
    ) -> float:
        """Cross-validation for a single outcome.

        Args:
            dataset: Training data for the model (for one outcome for
                the default `Surrogate`, with the exception of batched
                multi-output case, where training data is formatted with just
                one X and concatenated Ys).
            model_config: The model_config.
            search_space_digest: Search space digest used to set up model arguments.
            state_dict: Optional state dict to load.

        Returns:
            The eval criterion value for the given model config.
        """
        if isinstance(dataset, MultiTaskDataset):
            # only evaluate model on target task
            target_dataset = dataset.datasets[dataset.target_outcome_name]
        else:
            target_dataset = dataset
        X, Y = target_dataset.X, target_dataset.Y
        num_folds = self.surrogate_spec.num_folds
        if num_folds is None or num_folds > X.shape[0]:
            num_folds = X.shape[0]
        test_folds = np.array_split(
            ary=torch.arange(X.shape[0], device=X.device),
            indices_or_sections=num_folds,
        )
        cv_folds = (
            get_cv_fold(dataset=dataset, X=X, Y=Y, idcs=idcs)  # pyre-ignore[6]
            for idcs in test_folds
        )

        pred_Y = []
        pred_Yvar = []
        obs_Y = []
        for fold in cv_folds:
            # Since each CV fold removes points from the training data, the
            # remaining observations will not pass the input scaling checks.
            # To avoid confusing users with warnings, we disable these checks.
            with validate_input_scaling(False):
                loo_model = self._construct_model(
                    dataset=fold.train_dataset,
                    search_space_digest=search_space_digest,
                    model_config=model_config,
                    state_dict=state_dict,
                    refit=self.refit_on_cv,
                )
            # evaluate model
            with torch.no_grad():
                posterior = loo_model.posterior(
                    fold.test_X,
                    observation_noise=self.surrogate_spec.use_posterior_predictive,
                )
                # TODO: support non-GPyTorch posteriors
                posterior = assert_is_instance(posterior, GPyTorchPosterior)
                if isinstance(posterior, GaussianMixturePosterior):
                    pred_mean = posterior.mixture_mean
                    pred_var = posterior.mixture_variance
                else:
                    pred_mean = posterior.mean
                    pred_var = posterior.variance
            pred_Y.append(pred_mean)
            pred_Yvar.append(pred_var)
            obs_Y.append(fold.test_Y)
        # Stack results
        pred_Y = torch.cat(pred_Y)
        pred_Yvar = torch.cat(pred_Yvar)
        obs_Y = torch.cat(obs_Y)
        # evaluate model fit metric
        diag_fn = DIAGNOSTIC_FNS[none_throws(self.surrogate_spec.eval_criterion)]
        return diag_fn(
            y_obs=obs_Y.view(-1).cpu().numpy(),
            y_pred=pred_Y.view(-1).cpu().numpy(),
            se_pred=np.sqrt(pred_Yvar.view(-1).cpu().numpy()),
        )

    def _discard_cached_model_and_data_if_search_space_digest_changed(
        self, search_space_digest: SearchSpaceDigest
    ) -> None:
        """Checks whether the search space digest has changed since the last call
        to `fit`. If it has, discards cached model and datasets. Also updates
        `self._last_search_space_digest` for future checks.
        """
        if (
            self._last_search_space_digest is not None
            and search_space_digest != self._last_search_space_digest
        ):
            logger.debug(
                "Discarding all previously trained models due to a change "
                "in the search space digest."
            )
            self._submodels = {}
            self._last_datasets = {}
            self.metric_to_best_model_config = {}
        self._last_search_space_digest = search_space_digest

    def predict(
        self, X: Tensor, use_posterior_predictive: bool = False
    ) -> tuple[Tensor, Tensor]:
        """Predicts outcomes given an input tensor.

        Args:
            X: A ``n x d`` tensor of input parameters.
            use_posterior_predictive: A boolean indicating if the predictions
                should be from the posterior predictive (i.e. including
                observation noise).

        Returns:
            Tensor: The predicted posterior mean as an ``n x o``-dim tensor.
            Tensor: The predicted posterior covariance as a ``n x o x o``-dim tensor.
        """
        return predict_from_model(
            model=self.model, X=X, use_posterior_predictive=use_posterior_predictive
        )

    def best_in_sample_point(
        self,
        search_space_digest: SearchSpaceDigest,
        torch_opt_config: TorchOptConfig,
        options: TConfig | None = None,
    ) -> tuple[Tensor, float]:
        """Finds the best observed point and the corresponding observed outcome
        values.
        """
        if torch_opt_config.is_moo:
            raise NotImplementedError(
                "Best observed point is incompatible with MOO problems."
            )
        best_point_and_observed_value = best_in_sample_point(
            Xs=self.Xs,
            model=self,
            bounds=search_space_digest.bounds,
            objective_weights=torch_opt_config.objective_weights,
            outcome_constraints=torch_opt_config.outcome_constraints,
            linear_constraints=torch_opt_config.linear_constraints,
            fixed_features=torch_opt_config.fixed_features,
            risk_measure=torch_opt_config.risk_measure,
            options=options,
        )
        if best_point_and_observed_value is None:
            raise ValueError("Could not obtain best in-sample point.")
        best_point, observed_value = best_point_and_observed_value
        return (
            best_point.to(dtype=self.dtype, device=torch.device("cpu")),
            observed_value,
        )

    def best_out_of_sample_point(
        self,
        search_space_digest: SearchSpaceDigest,
        torch_opt_config: TorchOptConfig,
        options: TConfig | None = None,
    ) -> tuple[Tensor, Tensor]:
        """Finds the best predicted point and the corresponding value of the
        appropriate best point acquisition function.

        Args:
            search_space_digest: A `SearchSpaceDigest`.
            torch_opt_config: A `TorchOptConfig`; none-None `fixed_features` is
                not supported.
            options: Optional. If present, `seed_inner` (default None) and `qmc`
                (default True) will be parsed from `options`; any other keys
                will be ignored.

        Returns:
            A two-tuple (`candidate`, `acqf_value`), where `candidate` is a 1d
            Tensor of the best predicted point and `acqf_value` is a scalar (0d)
            Tensor of the acquisition function value at the best point.
        """
        if torch_opt_config.fixed_features:
            # When have fixed features, need `FixedFeatureAcquisitionFunction`
            # which has peculiar instantiation (wraps another acquisition fn.),
            # so need to figure out how to handle.
            # TODO (ref: https://fburl.com/diff/uneqb3n9)
            raise NotImplementedError("Fixed features not yet supported.")

        options = options or {}
        botorch_acqf_class, botorch_acqf_options = (
            pick_best_out_of_sample_point_acqf_class(
                outcome_constraints=torch_opt_config.outcome_constraints,
                seed_inner=assert_is_instance_optional(
                    options.get(Keys.SEED_INNER, None), int
                ),
                qmc=assert_is_instance(
                    options.get(Keys.QMC, True),
                    bool,
                ),
                risk_measure=torch_opt_config.risk_measure,
            )
        )

        # Avoiding circular import between `Surrogate` and `Acquisition`.
        from ax.generators.torch.botorch_modular.acquisition import Acquisition

        acqf = Acquisition(
            surrogate=self,
            botorch_acqf_class=botorch_acqf_class,
            search_space_digest=search_space_digest,
            torch_opt_config=torch_opt_config,
            botorch_acqf_options=botorch_acqf_options,
        )
        candidates, acqf_value, _ = acqf.optimize(
            n=1,
            search_space_digest=search_space_digest,
            inequality_constraints=_to_inequality_constraints(
                linear_constraints=torch_opt_config.linear_constraints
            ),
            fixed_features=torch_opt_config.fixed_features,
        )
        return candidates[0], acqf_value

    def pareto_frontier(self) -> tuple[Tensor, Tensor]:
        """For multi-objective optimization, retrieve Pareto frontier instead
        of best point.

        Returns: A two-tuple of:
            - tensor of points in the feature space,
            - tensor of corresponding (multiple) outcomes.
        """
        raise NotImplementedError("Pareto frontier not yet implemented.")

    def compute_diagnostics(self) -> dict[str, Any]:
        """Computes model diagnostics like cross-validation measure of fit, etc."""
        return {}

    def _serialize_attributes_as_kwargs(self) -> dict[str, Any]:
        """Serialize attributes of this surrogate, to be passed back to it
        as kwargs on reinstantiation.
        """
        return {
            "surrogate_spec": self.surrogate_spec,
            "refit_on_cv": self.refit_on_cv,
            "warm_start_refit": self.warm_start_refit,
            "metric_to_best_model_config": self.metric_to_best_model_config,
        }

    @property
    def outcomes(self) -> list[str]:
        if self._outcomes is None:
            raise ModelError("`outcomes` was not initialized. Please call `fit` first.")
        return self._outcomes

    @outcomes.setter
    def outcomes(self, value: list[str]) -> None:
        raise RuntimeError("Setting outcomes manually is disallowed.")

    @property
    def model_name_by_metric(self) -> dict[str, str]:
        """Returns a dictionary mapping metric names to model names."""
        return {
            metric_name: model_config.identifier
            for metric_name, model_config in (self.metric_to_best_model_config.items())
        }

    def models_for_gen(self, n: int) -> tuple[list[dict[str, str]], list[Model]]:
        """For each metric, draw `n` samples of models to use for generating `n`
        candidates.

        This method samples which models to use proportional to their model
        selection eval criterion value. Sampling is made scale-free with
        respect to the eval criterion by scaling with respect to the minimum
        value across models. This means the model with worst eval criterion
        will never be sampled, and other models will be sampled proportional
        to the amount of the gap to best value they are able to capture. This
        also means that sampling requires at least 3 models, and if fitting
        here used less than 3, the default self.model will be returned. In that
        case, the method will return a list with just the 1 model and not n.

        The output length will thus be either 1 or n, depending on if the method
        was successful in sampling n models or not.

        Args:
            n: The number of models to return.

        Returns:
            model_names: A list of the model_name_by_metric dict used for each
                candidate.
            models: A list of the models to be used for each candidate.
        """
        if len(self._model_name_to_eval[self.outcomes[0]]) < 3:
            logger.debug("Less than 3 models, no sampling to do for batch generation.")
            return [self.model_name_by_metric], [self.model]

        models = []
        model_names = []
        maximize = (
            MODEL_SELECTION_METRIC_DIRECTIONS[self.surrogate_spec.eval_criterion]
            == ModelFitMetricDirection.MAXIMIZE
        )
        for _ in range(n):
            models_i = []
            model_names_i = {}
            for outcome in self.outcomes:
                w = np.array(list(self._model_name_to_eval[outcome].values()))
                if not maximize:
                    w = -w
                w = w - w.min()
                if w.max() < 1e-10:
                    # All eval values are identical within tolerance.
                    # Uniform sampling.
                    w = np.ones_like(w)
                w = w / w.sum()
                model_name = str(
                    np.random.choice(
                        list(self._model_name_to_eval[outcome].keys()), p=w, size=1
                    )[0]
                )
                models_i.append(self._model_name_to_model[outcome][model_name])
                model_names_i[outcome] = model_name
            if isinstance(self._model, ModelListGP):
                models.append(ModelListGP(*models_i))
            elif isinstance(self._model, ModelList):
                models.append(ModelList(*models_i))
            elif len(models_i) > 1:
                # If MBM supports ModelList in the future, this will need to be
                # updated here.
                raise ValueError(
                    "Got multiple models but not a ModelListGP or ModelList."
                )  # pragma: no cover
            else:
                models.append(models_i[0])
            model_names.append(model_names_i)
        return model_names, models


submodel_input_constructor = Dispatcher(
    name="submodel_input_constructor", encoder=_argparse_type_encoder
)


@submodel_input_constructor.register(Model)
def _submodel_input_constructor_base(
    botorch_model_class: type[Model],
    model_config: ModelConfig,
    dataset: SupervisedDataset,
    search_space_digest: SearchSpaceDigest,
    surrogate: Surrogate,
) -> dict[str, Any]:
    """Construct the inputs required to initialize a BoTorch model.

    Args:
        botorch_model_class: The BoTorch model class to instantiate.
        model_config: The model config.
        dataset: The training data for the model.
        search_space_digest: Search space digest used to set up model arguments.
        surrogate: A reference to the surrogate that created the model.
            This can be used by the constructor to obtain any additional
            arguments that are not readily available.

    Returns:
        A dictionary of inputs for constructing the model.
    """
    model_kwargs_from_ss = _extract_model_kwargs(
        search_space_digest=search_space_digest, botorch_model_class=botorch_model_class
    )
    # pop task feature if this is not a multi-task model
    task_feature = model_kwargs_from_ss.get("task_feature", None)
    if not issubclass(botorch_model_class, MultiTaskGP):
        if task_feature is not None:
            model_kwargs_from_ss.pop("task_feature")
            logger.debug(
                f"Removing task feature for {botorch_model_class.__name__}",
                stacklevel=5,
            )
    formatted_model_inputs: dict[str, Any] = botorch_model_class.construct_inputs(
        training_data=dataset,
        **model_config.model_options,
        **model_kwargs_from_ss,
    )
    submodules = _construct_submodules(
        model_config=model_config,
        dataset=dataset,
        # This is used when constructing the input transforms.
        search_space_digest=search_space_digest,
        # Used to check for supported arguments and in covar module input constructors.
        botorch_model_class=botorch_model_class,
    )
    formatted_model_inputs.update(submodules)
    return formatted_model_inputs


@submodel_input_constructor.register(MultiTaskGP)
def _submodel_input_constructor_mtgp(
    botorch_model_class: type[Model],
    model_config: ModelConfig,
    dataset: SupervisedDataset,
    search_space_digest: SearchSpaceDigest,
    surrogate: Surrogate,
) -> dict[str, Any]:
    if len(dataset.outcome_names) > 1:
        raise NotImplementedError("Multi-output Multi-task GPs are not yet supported.")
    formatted_model_inputs = _submodel_input_constructor_base(
        botorch_model_class=botorch_model_class,
        model_config=model_config,
        dataset=dataset,
        search_space_digest=search_space_digest,
        surrogate=surrogate,
    )
    task_feature = formatted_model_inputs.get("task_feature")
    if task_feature is None:
        return formatted_model_inputs
    # specify output tasks so that model.num_outputs = 1
    # since the model only models a single outcome
    if formatted_model_inputs.get("output_tasks") is None:
        # SSD doesn't use -1, so we need to normalize here
        task_feature = none_throws(
            normalize_indices(indices=[task_feature], d=len(dataset.feature_names))
        )[0]
        if (search_space_digest.target_values is not None) and (
            target_value := search_space_digest.target_values.get(task_feature)
        ) is not None:
            formatted_model_inputs["output_tasks"] = [int(target_value)]
        else:
            raise UserInputError(
                "output_tasks or target task value must be provided for MultiTaskGP."
            )
    return formatted_model_inputs
